"""
An experimental implementation of S3 in stor
"""
from functools import partial
import logging
from multiprocessing.pool import ThreadPool
import os
import tempfile
import threading

import boto3
from boto3 import exceptions as boto3_exceptions
from boto3.s3.transfer import S3Transfer
from boto3.s3.transfer import TransferConfig
from botocore import exceptions as botocore_exceptions
import six

from stor import exceptions
from stor import settings
from stor import utils
from stor.base import Path
from stor.obs import OBSFile
from stor.obs import OBSPath
from stor.obs import OBSUploadObject

# Thread-local variable used to cache the client
_thread_local = threading.local()

logger = logging.getLogger(__name__)
progress_logger = logging.getLogger('%s.progress' % __name__)

S3File = OBSFile


def _parse_s3_error(exc, **kwargs):
    """
    Parses botocore.exception.ClientError exceptions to throw a more
    informative exception.
    """
    http_status = exc.response.get('ResponseMetadata', {}).get('HTTPStatusCode')
    msg = exc.response['Error'].get('Message', 'Unknown')

    # Give some more info about the error's origins
    if 'Bucket' in kwargs:
        msg += ' Bucket: ' + kwargs.get('Bucket')
    if 'Key' in kwargs:
        msg += ', Key: ' + kwargs.get('Key')

    if http_status == 403:
        return exceptions.UnauthorizedError(msg, exc)
    elif http_status == 404:
        return exceptions.NotFoundError(msg, exc)
    elif http_status == 503:
        return exceptions.UnavailableError(msg, exc)

    return exceptions.RemoteError(msg, exc)


def _get_s3_client():
    """Returns the boto3 client and initializes one if it doesn't already exist.

    We use a different boto3 client for each thread/process because boto3 clients
    are not thread-safe.

    Returns:
        boto3.Client: An instance of the S3 client.
    """
    if not hasattr(_thread_local, 's3_client'):
        session = boto3.session.Session()
        _thread_local.s3_client = session.client('s3')
    return _thread_local.s3_client


def _get_s3_transfer(config=None):
    """Returns a boto3 S3Transfer object and initializes one if it doesn't
    already exist or if config options are different.

    Like the boto3 client, we use a different transfer for each thread/process
    for thread-safety.

    Args:
        config (dict): A dict of config options

    Returns:
        boto3.s3.S3Transfer: An instance of an S3Transfer object.
    """
    if (not hasattr(_thread_local, 's3_transfer') or
            getattr(_thread_local, 's3_transfer_config', None) != config):
        transfer_config = None
        if config:
            transfer_config = TransferConfig(**config)
        _thread_local.s3_transfer = S3Transfer(_get_s3_client(), config=transfer_config)
        _thread_local.s3_transfer_config = config
    return _thread_local.s3_transfer


class S3DownloadLogger(utils.BaseProgressLogger):
    def __init__(self, total_download_objects):
        super(S3DownloadLogger, self).__init__(progress_logger)
        self.total_download_objects = total_download_objects
        self.downloaded_bytes = 0

    def update_progress(self, result):
        """Tracks number of bytes downloaded."""
        self.downloaded_bytes += (os.path.getsize(result['dest'])
                                  if not utils.has_trailing_slash(result['source']) else 0)

    def get_start_message(self):
        return 'starting download of %s objects' % self.total_download_objects

    def get_finish_message(self):
        return 'download complete - %s' % self.get_progress_message()

    def get_progress_message(self):
        elapsed_time = self.get_elapsed_time()
        formatted_elapsed_time = self.format_time(elapsed_time)
        mb = self.downloaded_bytes / (1024 * 1024.0)
        mb_s = mb / elapsed_time.total_seconds() if elapsed_time else 0.0
        return (
            '%s/%s\t'
            '%s\t'
            '%0.2f MB\t'
            '%0.2f MB/s'
        ) % (self.num_results, self.total_download_objects, formatted_elapsed_time, mb, mb_s)


class S3UploadLogger(utils.BaseProgressLogger):
    def __init__(self, total_upload_objects):
        super(S3UploadLogger, self).__init__(progress_logger)
        self.total_upload_objects = total_upload_objects
        self.uploaded_bytes = 0

    def update_progress(self, result):
        """Keep track of total uploaded bytes by referencing the object sizes"""
        self.uploaded_bytes += (os.path.getsize(result['source'])
                                if not utils.has_trailing_slash(result['dest']) else 0)

    def get_start_message(self):
        return 'starting upload of %s objects' % self.total_upload_objects

    def get_finish_message(self):
        return 'upload complete - %s' % self.get_progress_message()

    def get_progress_message(self):
        elapsed_time = self.get_elapsed_time()
        formatted_elapsed_time = self.format_time(elapsed_time)
        mb = self.uploaded_bytes / (1024 * 1024.0)
        mb_s = mb / elapsed_time.total_seconds() if elapsed_time else 0.0
        return (
            '%s/%s\t'
            '%s\t'
            '%0.2f MB\t'
            '%0.2f MB/s'
        ) % (self.num_results, self.total_upload_objects, formatted_elapsed_time, mb, mb_s)


class S3Path(OBSPath):
    """
    Provides the ability to manipulate and access S3 resources
    with a similar interface to the path library.

    Right now, the client defaults to Amazon S3 endpoints, but in the
    near-future, users should be able to custom configure the S3 client.

    Note that any S3 object whose name ends with a ``/`` is considered to be an empty directory
    marker.
    These objects will not be downloaded and instead an empty directory will be created.
    This follows Amazon's convention as described in the
    `S3 User Guide <http://docs.aws.amazon.com/AmazonS3/latest/UG/FolderOperations.html>`_.
    """
    drive = 's3://'

    @property
    def bucket(self):
        """Returns the bucket name from the path or None"""
        parts = self._get_parts()
        return parts[0] if len(parts) > 0 and parts[0] else None

    def _s3_client_call(self, method_name, *args, **kwargs):
        """
        Creates a boto3 S3 ``Client`` object and runs ``method_name``.
        """
        s3_client = _get_s3_client()
        method = getattr(s3_client, method_name)
        try:
            return method(*args, **kwargs)
        except botocore_exceptions.ClientError as e:
            six.raise_from(_parse_s3_error(e, **kwargs), e)

    def _get_s3_iterator(self, method_name, *args, **kwargs):
        """
        Creates a boto3 ``S3.Paginator`` object and returns an iterator
        that runs over results from ``method_name``.
        """
        s3_client = _get_s3_client()
        paginator = s3_client.get_paginator(method_name)
        return paginator.paginate(**kwargs)

    def _make_s3_transfer(self, method_name, config=None, *args, **kwargs):
        """
        Creates a boto3 ``S3Transfer`` object for doing multipart uploads
        and downloads and executes the given method.
        """
        transfer = _get_s3_transfer(config=config)
        method = getattr(transfer, method_name)
        try:
            return method(*args, **kwargs)
        except boto3_exceptions.S3UploadFailedError as e:
            six.raise_from(exceptions.FailedUploadError(str(e), e), e)
        except boto3_exceptions.RetriesExceededError as e:
            six.raise_from(exceptions.FailedDownloadError(str(e), e), e)

    def open(self, mode='r'):
        """
        Opens a S3File that can be read or written to.

        For examples of reading and writing opened objects, view
        S3File.

        Args:
            mode (str): The mode of object IO. Currently supports reading
                ("r" or "rb") and writing ("w", "wb")

        Returns:
            S3File: The s3 object.

        Raises:
            RemoteError: A s3 client error occurred.
        """
        return S3File(self, mode=mode)

    def list(self,
             starts_with=None,
             limit=None,
             condition=None,
             use_manifest=False,
             # hidden args
             list_as_dir=False,
             ignore_dir_markers=False):
        """
        List contents using the resource of the path as a prefix.

        Args:
            starts_with (str): Allows for an additional search path to be
                appended to the current swift path. The current path will be
                treated as a directory.
            limit (int): Limit the amount of results returned.
            condition (function(results) -> bool): The method will only return
                when the results matches the condition.
            use_manifest (bool): Perform the list and use the data manfest file to validate
                the list.

        Returns:
            List[S3Path]: Every path in the listing

        Raises:
            RemoteError: An s3 client error occurred.
            ConditionNotMetError: Results were returned, but they did not meet the condition.
        """
        bucket = self.bucket
        prefix = self.resource
        utils.validate_condition(condition)

        if use_manifest:
            object_names = utils.get_data_manifest_contents(self)
            manifest_cond = partial(utils.validate_manifest_list, object_names)
            condition = (utils.join_conditions(condition, manifest_cond)
                         if condition else manifest_cond)

        if starts_with:
            prefix = prefix / starts_with if prefix else starts_with
        else:
            prefix = prefix or ''

        list_kwargs = {
            'Bucket': bucket,
            'Prefix': prefix,
            'PaginationConfig': {}
        }

        if limit:
            list_kwargs['PaginationConfig']['MaxItems'] = limit

        if list_as_dir:
            # Ensure the the prefix has a trailing slash if there is a prefix
            list_kwargs['Prefix'] = utils.with_trailing_slash(prefix) if prefix else ''
            list_kwargs['Delimiter'] = '/'

        path_prefix = S3Path('%s%s' % (self.drive, bucket))

        results = self._get_s3_iterator('list_objects_v2', **list_kwargs)
        list_results = []
        try:
            for page in results:
                if 'Contents' in page:
                    list_results.extend([
                        path_prefix / result['Key']
                        for result in page['Contents']
                        if not ignore_dir_markers or
                        (ignore_dir_markers and not utils.has_trailing_slash(result['Key']))
                    ])
                if list_as_dir and 'CommonPrefixes' in page:
                    list_results.extend([
                        path_prefix / result['Prefix']
                        for result in page['CommonPrefixes']
                    ])
        except botocore_exceptions.ClientError as e:
            six.raise_from(_parse_s3_error(e), e)

        utils.check_condition(condition, list_results)
        return list_results

    def listdir(self):
        """List the path as a dir, returning top-level directories and files."""
        return self.list(list_as_dir=True)

    def exists(self):
        """
        Checks existence of the path.

        Returns:
            bool: True if the path exists, False otherwise.

        Raises:
            RemoteError: A non-404 error occurred.
        """
        if not self.resource:
            try:
                return bool(self._s3_client_call('head_bucket', Bucket=self.bucket))
            except exceptions.NotFoundError:
                return False
        try:
            return bool(self.stat())
        except exceptions.NotFoundError:
            pass
        try:
            return bool(utils.with_trailing_slash(self).list(limit=1))
        except exceptions.NotFoundError:
            return False

    def isdir(self):
        """
        Any S3 object whose name ends with a ``/`` is considered to be an empty directory marker.
        These objects will not be downloaded and instead an empty directory will be created.
        This follows Amazon's convention as described in the
        `S3 User Guide <http://docs.aws.amazon.com/AmazonS3/latest/UG/FolderOperations.html>`_.
        """
        # Handle buckets separately (in case the bucket is empty)
        if not self.resource:
            try:
                return bool(self._s3_client_call('head_bucket', Bucket=self.bucket))
            except exceptions.NotFoundError:
                return False
        try:
            return bool(utils.with_trailing_slash(self).list(limit=1))
        except exceptions.NotFoundError:
            return False

    def isfile(self):
        try:
            return self.stat() and not utils.has_trailing_slash(self)
        except (exceptions.NotFoundError, ValueError):
            return False

    def getsize(self):
        """
        Returns the content length of an object in S3.

        Directories and buckets have no length and will return 0.
        """
        bucket = self.bucket
        if not self.resource:
            # check for existence of bucket
            self._s3_client_call('head_bucket', Bucket=bucket)
        else:
            try:
                return self._s3_client_call('head_object',
                                            Bucket=bucket,
                                            Key=self.resource).get('ContentLength', 0)
            except exceptions.NotFoundError:
                # Check if path is a directory
                if not self.exists():
                    raise
        return 0

    def remove(self):
        """
        Removes a single object.

        Raises:
            ValueError: The path is invalid.
            RemoteError: An s3 client error occurred.
        """
        resource = self.resource
        if not resource:
            raise ValueError('cannot remove a bucket')
        return self._s3_client_call('delete_object', Bucket=self.bucket, Key=resource)

    def rmtree(self):
        """
        Removes a resource and all of its contents. The path should point to a directory.

        If the specified resource is an object, nothing will happen.
        """
        # Ensure there is a trailing slash (path is a dir)
        delete_path = utils.with_trailing_slash(self)
        delete_list = delete_path.list()

        while len(delete_list) > 0:
            # boto3 only allows deletion of up to 1000 objects at a time
            len_range = min(len(delete_list), 1000)
            objects = {
                'Objects': [
                    {'Key': delete_list.pop(0).resource}
                    for i in range(len_range)
                ]
            }
            response = self._s3_client_call('delete_objects', Bucket=self.bucket, Delete=objects)

            if 'Errors' in response:
                raise exceptions.RemoteError('an error occurred while using rmtree: %s, Key: %s'
                                             % (response['Errors'][0].get('Message'),
                                                response['Errors'][0].get('Key')),
                                             response['Errors'])

    def stat(self):
        """
        Performs a stat on the path.

        ``stat`` only works on paths that are objects.
        Using ``stat`` on a directory of objects will produce a `NotFoundError`.

        An example return dictionary is the following::

            {
                'DeleteMarker': True|False,
                'AcceptRanges': 'string',
                'Expiration': 'string',
                'Restore': 'string',
                'LastModified': datetime(2015, 1, 1),
                'ContentLength': 123,
                'ETag': 'string',
                'MissingMeta': 123,
                'VersionId': 'string',
                'CacheControl': 'string',
                'ContentDisposition': 'string',
                'ContentEncoding': 'string',
                'ContentLanguage': 'string',
                'ContentType': 'string',
                'Expires': datetime(2015, 1, 1),
                'WebsiteRedirectLocation': 'string',
                'ServerSideEncryption': 'AES256'|'aws:kms',
                'Metadata': {
                    'string': 'string'
                },
                'SSECustomerAlgorithm': 'string',
                'SSECustomerKeyMD5': 'string',
                'SSEKMSKeyId': 'string',
                'StorageClass': 'STANDARD'|'REDUCED_REDUNDANCY'|'STANDARD_IA',
                'RequestCharged': 'requester',
                'ReplicationStatus': 'COMPLETE'|'PENDING'|'FAILED'|'REPLICA'
            }
        """
        if not self.resource:
            raise ValueError('stat cannot be called on a bucket')

        response = self._s3_client_call('head_object', Bucket=self.bucket, Key=self.resource)
        response = {
            key: val for key, val in response.items()
            if key is not 'ResponseMetadata'
        }
        return response

    def read_object(self):
        """"Reads an individual object."""
        body = self._s3_client_call('get_object', Bucket=self.bucket, Key=self.resource)['Body']
        return body.read()

    def write_object(self, content):
        """Writes an individual object.

        Note that this method writes the provided content to a temporary
        file before uploading.

        Args:
            content (str): The content of the object
        """
        mode = 'wb' if isinstance(content, bytes) else 'w'
        with tempfile.NamedTemporaryFile(mode) as fp:
            fp.write(content)
            fp.flush()
            self.upload([OBSUploadObject(fp.name, self.resource)])

    def download_object(self, dest, config=None, **kwargs):
        """
        Downloads a file from S3 to a destination file.

        Args:
            dest (str): The destination path to download file to.

        Notes:
            - The destination directory will be created automatically if it doesn't exist.

            - This method downloads to paths relative to the current
              directory.
        """
        result = {
            'source': self,
            'dest': dest,
            'success': True
        }
        if utils.has_trailing_slash(self):
            # Handle directory markers separately
            utils.make_dest_dir(str(dest))
            return result

        dl_kwargs = {
            'bucket': self.bucket,
            'key': str(self.resource),
            'filename': str(dest),
            'config': config
        }
        utils.make_dest_dir(self.parts_class(dest).parent)
        try:
            self._make_s3_transfer('download_file', **dl_kwargs)
        except exceptions.RemoteError as e:
            result['success'] = False
            result['error'] = e
        return result

    def _download_object_worker(self, obj_params, config=None):
        """Downloads a single object. Helper for threaded download."""
        name = self.parts_class(obj_params['source'][len(utils.with_trailing_slash(self)):])
        return obj_params['source'].download_object(obj_params['dest'] / name, config=config)

    def download(self, dest, condition=None, use_manifest=False, **kwargs):
        """Downloads a directory from S3 to a destination directory.

        Args:
            dest (str): The destination path to download file to. If downloading to a directory,
                there must be a trailing slash. The directory will be created if it doesn't exist.
            condition (function(results) -> bool): The method will only return
                when the results of download matches the condition.

        Returns:
            List[S3Path]: A list of the downloaded objects.

        Notes:
        - The destination directory will be created automatically if it doesn't exist.
        - This method downloads to paths relative to the current directory.
        """
        utils.validate_condition(condition)

        if use_manifest:
            object_names = utils.get_data_manifest_contents(self)
            manifest_cond = partial(utils.validate_manifest_list, object_names)
            condition = (utils.join_conditions(condition, manifest_cond)
                         if condition else manifest_cond)

        source = utils.with_trailing_slash(self)
        files_to_download = [
            {'source': file, 'dest': dest}
            for file in source.list()
        ]

        options = settings.get()['s3:download']
        segment_size = utils.str_to_bytes(options.get('segment_size'))
        transfer_config = {
            'multipart_threshold': segment_size,
            'max_concurrency': options.get('segment_threads'),
            'multipart_chunksize': segment_size
        }
        download_w_config = partial(self._download_object_worker, config=transfer_config)

        downloaded = {'completed': [], 'failed': []}
        with S3DownloadLogger(len(files_to_download)) as dl:
            pool = ThreadPool(options['object_threads'])
            try:
                result_iter = pool.imap_unordered(download_w_config, files_to_download)
                while True:
                    try:
                        result = result_iter.next(0xFFFF)
                        if result['success']:
                            dl.add_result(result)
                            downloaded['completed'].append(result)
                        else:
                            downloaded['failed'].append(result)
                    except StopIteration:
                        break
                pool.close()
            except:
                pool.terminate()
                raise
            finally:
                pool.join()

        if downloaded['failed']:
            raise exceptions.FailedDownloadError('an error occurred while downloading', downloaded)

        utils.check_condition(condition, [r['source'] for r in downloaded['completed']])
        return downloaded

    def _upload_object(self, upload_obj, config=None):
        """Upload a single object given an OBSUploadObject."""
        if utils.has_trailing_slash(upload_obj.object_name):
            # Handle empty directories separately
            ul_kwargs = {
                'Bucket': self.bucket,
                'Key': utils.with_trailing_slash(str(upload_obj.object_name))
            }
            if upload_obj.options and 'headers' in upload_obj.options:
                ul_kwargs.update(upload_obj.options['headers'])
            s3_call = self._s3_client_call
            method = 'put_object'
        else:
            ul_kwargs = {
                'bucket': self.bucket,
                'key': str(upload_obj.object_name),
                'filename': upload_obj.source,
                'config': config
            }
            if upload_obj.options and 'headers' in upload_obj.options:
                ul_kwargs['extra_args'] = upload_obj.options['headers']
            s3_call = self._make_s3_transfer
            method = 'upload_file'

        result = {
            'source': upload_obj.source,
            'dest': S3Path(self.drive + self.bucket) / (ul_kwargs.get('key') or
                                                        ul_kwargs.get('Key')),
            'success': True
        }

        try:
            s3_call(method, **ul_kwargs)
        except exceptions.RemoteError as e:
            result['success'] = False
            result['error'] = e

        return result

    def upload(self, source, condition=None, use_manifest=False, headers=None, **kwargs):
        """Uploads a list of files and directories to s3.

        Note that the S3Path is treated as a directory.

        Note that for user-provided OBSUploadObjects, an empty directory's destination
        must have a trailing slash.

        Args:
            source (List[str|OBSUploadObject]): A list of source files, directories, and
                OBSUploadObjects to upload to S3.
            condition (function(results) -> bool): The method will only return
                when the results of upload matches the condition.
            use_manifest (bool): Generate a data manifest and validate the upload results
                are in the manifest.
            headers (dict): A dictionary of object headers to apply to the object.
                Headers will not be applied to OBSUploadObjects and any headers
                specified by an OBSUploadObject will override these headers.
                Headers should be specified as key-value pairs,
                e.g. {'ContentLanguage': 'en'}

        Returns:
            List[S3Path]: A list of the uploaded files as S3Paths.

        Notes:

        - This method uploads to paths relative to the current
          directory.
        """
        if use_manifest and not (len(source) == 1 and os.path.isdir(source[0])):
            raise ValueError('can only upload one directory with use_manifest=True')
        utils.validate_condition(condition)

        files_to_convert = utils.walk_files_and_dirs([
            name for name in source if not isinstance(name, OBSUploadObject)
        ])
        files_to_upload = [
            obj for obj in source if isinstance(obj, OBSUploadObject)
        ]

        manifest_file_name = (Path(source[0]) / utils.DATA_MANIFEST_FILE_NAME
                              if use_manifest else None)
        resource_base = self.resource or Path('')
        files_to_upload.extend([
            OBSUploadObject(
                name,
                resource_base / (utils.with_trailing_slash(utils.file_name_to_object_name(name))
                                 if Path(name).isdir() else utils.file_name_to_object_name(name)),
                options={'headers': headers} if headers else None)
            for name in files_to_convert if name != manifest_file_name
        ])

        if use_manifest:
            # Generate the data manifest and save it remotely
            object_names = [o.object_name for o in files_to_upload]
            utils.generate_and_save_data_manifest(source[0], object_names)
            manifest_obj_name = resource_base / utils.file_name_to_object_name(manifest_file_name)
            manifest_obj = OBSUploadObject(str(manifest_file_name),
                                           manifest_obj_name,
                                           options={'headers': headers} if headers else None)
            self._upload_object(manifest_obj)

            # Make a condition for validating the upload
            manifest_cond = partial(utils.validate_manifest_list, object_names)
            condition = (utils.join_conditions(condition, manifest_cond)
                         if condition else manifest_cond)

        options = settings.get()['s3:upload']
        segment_size = utils.str_to_bytes(options.get('segment_size'))
        transfer_config = {
            'multipart_threshold': segment_size,
            'max_concurrency': options.get('segment_threads'),
            'multipart_chunksize': segment_size
        }
        upload_w_config = partial(self._upload_object, config=transfer_config)

        uploaded = {'completed': [], 'failed': []}
        with S3UploadLogger(len(files_to_upload)) as ul:
            pool = ThreadPool(options['object_threads'])
            try:
                result_iter = pool.imap_unordered(upload_w_config, files_to_upload)
                while True:
                    try:
                        result = result_iter.next(0xFFFF)
                        if result['success']:
                            ul.add_result(result)
                            uploaded['completed'].append(result)
                        else:
                            uploaded['failed'].append(result)
                    except StopIteration:
                        break
                pool.close()
            except:
                pool.terminate()
                raise
            finally:
                pool.join()

        if uploaded['failed']:
            raise exceptions.FailedUploadError('an error occurred while uploading', uploaded)

        utils.check_condition(condition, [r['dest'] for r in uploaded['completed']])
        return uploaded



import errno
import fnmatch
import glob
import os
import ntpath
import posixpath
import shutil
import sys
import warnings

from six.moves import builtins
from six import text_type
from six import string_types
from six import PY3

from stor import utils


class TreeWalkWarning(Warning):
    pass


class Path(text_type):
    """
    Wraps path operations with an object-oriented API that makes it easier to
    combine and also to work with OBS and local paths via a single API. Methods
    on this class will be implemented by all subclasses of path.

    Using the class-level constructor returns a concrete subclass based on
    prefix and current environment.

    Examples::

        >>> from stor import Path
        >>> Path('/some/path')
        PosixPath('/some/path')
        >>> Path('swift://AUTH_something/cont/blah')
        SwiftPath('swift://AUTH_something/cont/blah')
        >>> Path('s3://bucket/prefix/key')
        S3Path('s3://bucket/prefix/key')
    """

    def __new__(cls, path):
        if cls is Path:
            if not hasattr(path, 'startswith'):
                raise TypeError('must be a string like')
            if utils.is_swift_path(path):
                from stor.swift import SwiftPath

                cls = SwiftPath
            elif utils.is_s3_path(path):
                from stor.s3 import S3Path

                cls = S3Path
            elif os.path == ntpath:
                from stor.windows import WindowsPath

                cls = WindowsPath
            elif os.path == posixpath:
                from stor.posix import PosixPath

                cls = PosixPath
            else:  # pragma: no cover
                assert False, 'path is not compatible with stor'
        return text_type.__new__(cls, path)

    def __init__(self, path):
        super(Path, self).__init__()

    @utils.ClassProperty
    @classmethod
    def path_class(cls):
        """What class should be used to construct new instances from this class"""
        return cls

    @utils.ClassProperty
    @classmethod
    def parts_class(cls):
        """What class should be used to construct path *components*"""
        return cls

    def _has_incompatible_path_module(self, other):
        """Returns true if the other path is a stor path and has a
        compatible path module for path operations"""
        return isinstance(other, Path) and other.path_module != self.path_module

    copy = utils.copy
    copytree = utils.copytree

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, super(Path,
                           self).__repr__().lstrip('u'))

    def __div__(self, rel):
        """Join two path components (self / rel), adding a separator character if needed."""
        if self._has_incompatible_path_module(rel):
            return NotImplemented
        return self.path_class(self.path_module.join(self, rel))

    def __rdiv__(self, rel):
        """Join two path components (rel / self), adding a separator character if needed."""
        if self._has_incompatible_path_module(rel):
            return NotImplemented
        return self.path_class(self.path_module.join(rel, self))

    # Make the / operator work even when true division is enabled.
    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def __add__(self, more):
        if self._has_incompatible_path_module(more):
            return NotImplemented
        return self.path_class(super(Path, self).__add__(more))

    def __radd__(self, other):
        if self._has_incompatible_path_module(other):
            return NotImplemented
        if not isinstance(other, string_types):
            return NotImplemented
        return self.path_class(other.__add__(self))

    #
    # --- Operations on Path strings.

    def abspath(self):
        """See :func:`os.path.abspath` """
        return self.path_class(self.path_module.abspath(self))

    def normcase(self):
        """See :func:`os.path.normcase` """
        return self.path_class(self.path_module.normcase(self))

    def normpath(self):
        """See :func:`os.path.normpath` """
        return self.path_class(self.path_module.normpath(self))

    def realpath(self):  # pragma: no cover (temporary)
        """See :func:`os.path.realpath` """
        return self.path_class(self.path_module.realpath(self))

    def expanduser(self):
        """See :func:`os.path.expanduser` """
        return self.path_class(self.path_module.expanduser(self))

    def expandvars(self):
        """See :func:`os.path.expandvars` """
        return self.path_class(self.path_module.expandvars(self))

    def dirname(self):
        """See :attr:`parent`, :func:`os.path.dirname` """
        return self.path_class(self.path_module.dirname(self))

    def basename(self):
        """See :attr:`name`, :func:`os.path.basename` """
        return self.parts_class(self.path_module.basename(self))

    def expand(self):
        """ Clean up a filename by calling :meth:`expandvars()`,
        :meth:`expanduser()`, and :meth:`normpath()` on it.

        This is commonly everything needed to clean up a filename
        read from a configuration file, for example.
        """
        return self.expandvars().expanduser().normpath()

    def fnmatch(self, pattern, normcase=None):
        """Return ``True`` if :attr:`name` matches the given ``pattern``.

        .. seealso:: :func:`fnmatch.fnmatch`

        Args:
            pattern (str): A filename pattern with wildcards,
                for example ``'*.py'``. If the pattern contains a `normcase`
                attribute, it is applied to the name and path prior to comparison.
            normcase (func, optional): A function used to normalize the pattern and
                filename before matching. Defaults to :meth:`self.module`, which defaults
                to :meth:`os.path.normcase`.

        """
        default_normcase = getattr(pattern, 'normcase', self.path_module.normcase)
        normcase = normcase or default_normcase
        name = normcase(self.name)
        pattern = normcase(pattern)
        return fnmatch.fnmatchcase(name, pattern)

    @property
    def parent(self):
        return self.dirname()

    @property
    def name(self):
        return self.basename()

    @property
    def namebase(self):
        """ The same as :attr:`name`, but with one file extension stripped off.

        For example,
        ``Path('/home/guido/python.tar.gz').name == 'python.tar.gz'``,
        but
        ``Path('/home/guido/python.tar.gz').namebase == 'python.tar'``.
        """
        base, ext = self.path_module.splitext(self.name)
        return base

    @property
    def drive(self):
        return self.splitdrive()[0]

    @property
    def ext(self):
        """ The file extension, for example ``'.py'``. """
        return self.splitext()[1]

    def splitpath(self):
        """ p.splitpath() -> Return ``(p.parent, p.name)``.

        (naming is to avoid colliding with str.split)

        See: :attr:`parent`, :attr:`name`, :func:`os.path.split`
        """
        parent, child = self.path_module.split(self)
        return self.path_class(parent), child

    def splitext(self):
        """ p.splitext() -> Return ``(p.stripext(), p.ext)``.

        Split the filename extension from this path and return
        the two parts.  Either part may be empty.

        The extension is everything from ``'.'`` to the end of the
        last path segment.  This has the property that if
        ``(a, b) == p.splitext()``, then ``a + b == p``.

        See: :func:`os.path.splitext`
        """
        filename, ext = self.path_module.splitext(self)
        return self.path_class(filename), ext

    def splitdrive(self):
        """ p.splitdrive() -> Return ``(p.drive, <the rest of p>)``.

        Split the drive specifier from this path.  If there is
        no drive specifier, :samp:`{p.drive}` is empty, so the return value
        is simply ``(Path(''), p)``.  This is always the case on Unix.

        See: :func:`os.path.splitdrive`
        """
        drive, rel = self.path_module.splitdrive(self)
        return self.path_class(drive), rel

    def joinpath(self, *others):
        """
        Join first to zero or more :class:`Path` components, adding a separator
        character (:samp:`{first}.module.sep`) if needed.  Returns a new
        instance of :samp:`{first}.path_class`.

        See: :func:`os.path.join`
        """
        return self.path_class(self.path_module.join(self, *others))

    def open(self, **kwargs):
        raise NotImplementedError

    def list(self, *args, **kwargs):
        """List all contents using the path as a prefix.

        Note: Skips broken symlinks."""
        raise NotImplementedError

    def listdir(self):
        raise NotImplementedError

    def glob(self, pattern):
        """ Glob for pattern relative to this directory.

        Note that Swift currently only supports a single trailing *"""
        raise NotImplementedError

    def exists(self):
        """Checks whether path exists on local filesystem or on swift.

        For directories on swift, checks whether directory sentinel exists or
        at least one subdirectory exists"""
        raise NotImplementedError

    def isabs(self):
        """ See: :func:`os.path.isabs`

        Always True with SwiftPath"""
        raise NotImplementedError

    def isdir(self):
        """ See: :func:`os.path.isdir` """
        raise NotImplementedError

    def isfile(self):
        """ See: :func:`os.path.isfile` """
        raise NotImplementedError

    def islink(self):
        """ See: :func:`os.path.islink`

        Always False on Swift."""
        raise NotImplementedError

    def ismount(self):
        """ See: :func:`os.path.ismount`

        Always True on Swift.
        """
        raise NotImplementedError

    def getsize(self):
        """Returns size, in bytes of path.

        For Swift containers and tenants, will return 0. For POSIX directories,
        returns an undefined value.

        Raises:
            os.error: if file does not exist or is inaccessible
            NotFoundError/UnauthorizedError: from swift
        """
        raise NotImplementedError

    def remove(self):
        """ Delete single path or object """
        raise NotImplementedError

    def rmtree(self):
        """Delete entire directory (or all paths starting with prefix).

        See shutil.rmtree"""
        raise NotImplementedError

    def walkfiles(self, pattern=None):
        """Iterate over files recursively.

        Args:
            pattern (str, optional): Limits the results to files
                with names that match the pattern.  For example,
                ``mydir.walkfiles('*.tmp')`` yields only files with the ``.tmp``
                extension.

        Returns:
            Iter[Path]: Files recursively under the path
        """
        raise NotImplementedError


class FileSystemPath(Path):
    """'Abstract' class implementing file-system specific operations.

    In particular: allows changing directory when used as contextmanager and
    wraps Python's builtin open() to be compatible with swift_upload_args.
    """
    def open(self, *args, **kwargs):
        """
        Opens a path and retains interface compatibility with
        `SwiftPath` by popping the unused ``swift_upload_args`` keyword
        argument.
        """
        kwargs.pop('swift_upload_kwargs', None)
        return builtins.open(self, *args, **kwargs)

    def __enter__(self):
        self._old_dir = os.getcwd()
        os.chdir(self)
        return self

    def __exit__(self, *_):
        os.chdir(self._old_dir)

    def chdir(self):
        os.chdir(self)

    # no stat() because we want to provide a cross-compatible API at some point

    @staticmethod
    def _always_unicode(path):  # pragma: no cover (OS-dependent)
        """
        Ensure the path as retrieved from a Python API, such as
        :func:`os.listdir`, is a proper Unicode string.
        """
        if PY3 or isinstance(path, text_type):
            return path
        return path.decode(sys.getfilesystemencoding(), 'surrogateescape')

    def listdir(self):
        """D.listdir() -> List of items in this directory.

        The elements of the list are Path objects.
        """
        return [self / child
                for child in map(self._always_unicode, os.listdir(self))]

    def glob(self, pattern):
        cls = self.path_class
        return [cls(s) for s in glob.glob(self / pattern)]

    def exists(self):
        """ See: :func:`os.path.exists` """
        return self.path_module.exists(self)

    def isabs(self):
        """ See: :func:`os.path.isabs` """
        return self.path_module.isabs(self)

    def isdir(self):
        """ See: :func:`os.path.isdir` """
        return self.path_module.isdir(self)

    def isfile(self):
        """ See: :func:`os.path.isfile` """
        return self.path_module.isfile(self)

    def islink(self):  # pragma: no cover (temporary)
        """ See: :func:`os.path.islink` """
        return self.path_module.islink(self)

    def ismount(self):  # pragma: no cover (temporary)
        """ See: :func:`os.path.ismount` """
        return self.path_module.ismount(self)

    def getsize(self):
        """ See: :func:`os.path.getsize` """
        return self.path_module.getsize(self)

    def remove(self):
        """ See: :func:`os.remove` """
        os.remove(self)
        return self

    rmtree = shutil.rmtree

    def makedirs(self, mode=0o777):
        """ See: :func:`os.makedirs` """
        os.makedirs(self, mode)
        return self

    def makedirs_p(self, mode=0o777):
        """ Like :meth:`makedirs`, but does not raise an exception if the
        directory already exists. """
        try:
            self.makedirs(mode)
        except OSError:
            _, e, _ = sys.exc_info()
            if e.errno != errno.EEXIST:  # pragma: no cover (temporary)
                raise
        return self

    def _splitall(self):  # pragma: no cover (temporary)
        r""" Return a list of the path components in this path.

        The first item in the list will be a Path.  Its value will be
        either :data:`os.curdir`, :data:`os.pardir`, empty, or the root
        directory of this path (for example, ``'/'`` or ``'C:\\'``).  The
        other items in the list will be strings.

        ``path.Path.joinpath(*result)`` will yield the original path.
        """
        parts = []
        loc = self
        while loc != os.curdir and loc != os.pardir:
            prev = loc
            loc, child = prev.splitpath()
            if loc == prev:
                break
            parts.append(child)
        parts.append(loc)
        parts.reverse()
        return parts

    def relpath(self, start='.'):
        """ Return this path as a relative path,
        based from `start`, which defaults to the current working directory.
        """
        cwd = self.path_class(start)
        return cwd.relpathto(self)

    def relpathto(self, dest):  # pragma: no cover (temporary)
        """ Return a relative path from `self` to `dest`.

        If there is no relative path from `self` to `dest`, for example if
        they reside on different drives in Windows, then this returns
        ``dest.abspath()``.
        """
        origin = self.abspath()
        dest = self.path_class(dest).abspath()

        orig_list = origin.normcase()._splitall()
        # Don't normcase dest!  We want to preserve the case.
        dest_list = dest._splitall()

        if orig_list[0] != self.path_module.normcase(dest_list[0]):
            # Can't get here from there.
            return dest

        # Find the location where the two paths start to differ.
        i = 0
        for start_seg, dest_seg in zip(orig_list, dest_list):
            if start_seg != self.path_module.normcase(dest_seg):
                break
            i += 1

        # Now i is the point where the two paths diverge.
        # Need a certain number of "os.pardir"s to work up
        # from the origin to the point of divergence.
        segments = [os.pardir] * (len(orig_list) - i)
        # Need to add the diverging part of dest_list.
        segments += dest_list[i:]
        if len(segments) == 0:
            # If they happen to be identical, use os.curdir.
            relpath = os.curdir
        else:
            relpath = self.path_module.join(*segments)
        return self.path_class(relpath)

    def mkdir(self, mode=0o777):
        """ .. seealso:: :func:`os.mkdir` """
        os.mkdir(self, mode)
        return self

    def mkdir_p(self, mode=0o777):
        """ Like :meth:`mkdir`, but does not raise an exception if the
        directory already exists. """
        try:
            self.mkdir(mode)
        except OSError:
            _, e, _ = sys.exc_info()
            if e.errno != errno.EEXIST:  # pragma: no cover (temporary)
                raise
        return self

    def rmdir(self):
        """ .. seealso:: :func:`os.rmdir` """
        os.rmdir(self)
        return self

    def rmdir_p(self):
        """ Like :meth:`rmdir`, but does not raise an exception if the
        directory is not empty or does not exist. """
        try:
            self.rmdir()
        except OSError:  # pragma: no cover (temporary)
            _, e, _ = sys.exc_info()
            if e.errno != errno.ENOTEMPTY and e.errno != errno.ENOENT:
                raise
        return self

    def walkfiles(self, pattern=None, errors='strict'):  # flake8: noqa pragma: no cover 
        """ D.walkfiles() -> iterator over files in D, recursively.
        The optional argument `pattern` limits the results to files
        with names that match the pattern.  For example,
        ``mydir.walkfiles('*.tmp')`` yields only files with the ``.tmp``
        extension.
        """
        if errors not in ('strict', 'warn', 'ignore'):
            raise ValueError("invalid errors parameter")

        try:
            childList = self.listdir()
        except Exception:
            if errors == 'ignore':
                return
            elif errors == 'warn':
                warnings.warn(
                    "Unable to list directory '%s': %s"
                    % (self, sys.exc_info()[1]),
                    TreeWalkWarning)
                return
            else:
                raise

        for child in childList:
            try:
                isfile = child.isfile()
                isdir = not isfile and child.isdir()
            except:
                if errors == 'ignore':
                    continue
                elif errors == 'warn':
                    warnings.warn(
                        "Unable to access '%s': %s"
                        % (self, sys.exc_info()[1]),
                        TreeWalkWarning)
                    continue
                else:
                    raise

            if isfile:
                if pattern is None or child.fnmatch(pattern):
                    yield child
            elif isdir:
                for f in child.walkfiles(pattern, errors):
                    yield f



# -*- coding: utf-8 -*-
#
# stor documentation build configuration file
import os
import subprocess


def get_version():
    """
    Extracts the version number from setup.py
    """
    proc = subprocess.Popen(['make', '--no-print-directory', 'version'], cwd='..',
                            stdout=subprocess.PIPE)
    return str(proc.communicate()[0].strip())


# -- General configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'toc'

# default role for "`" (makes it attempt to match against references within project)
default_role = 'any'

# General information about the project.
project = u'stor'
copyright = u'2015, Counsyl Inc'

# The short X.Y version.
version = get_version()
# The full version, including alpha/beta/rc tags.
release = version

exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

intersphinx_mapping = {
    'python': ('http://python.readthedocs.org/en/v2.7.2/', None),
}

# -- Options for HTML output ----------------------------------------------

html_theme = 'default'
#html_theme_path = []

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# Output file base name for HTML help builder.
htmlhelp_basename = 'stor-doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', 'stor.tex', u'stor Documentation',
     u'Counsyl Inc.', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'stor', u'stor Documentation',
     [u'Counsyl Inc.'], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'stor', u'stor Documentation',
     u'Counsyl Inc.', 'stor', 'A short description',
     'Miscellaneous'),
]



"""
The CLI can be accessed through the command ``stor``. For details on
valid subcommands, usage, and input options, refer to the ``--help`` / ``-h``
flag.

In addition to UNIX-like basic list, copy, and remove commands, the CLI also
has some features such as specifying a current working directory on an OBS service
(which allows for relative paths), ``cat``, and copying from ``stdin``.

The ``list`` command is different from the ``ls`` command. ``list`` recursively
lists all files and directories under a given path while ``ls`` lists the path
as a directory in a way that is similar to the UNIX command.

To copy or remove a tree, use the ``-r`` flag with ``cp`` or ``remove``.

Relative Paths
--------------

Using the CLI, the user can specify a current working directory on supported
OBS services (currently swift and s3) using the ``cd`` subcommand::

    $ stor cd s3://bucket
    $ stor cd swift://tenant/container

To check the current working directory, use the ``pwd`` subcommand::

    $ stor pwd
    s3://bucket
    swift://tenant/container

To clear the current working directory, use the ``clear`` subcommand::

    $ stor clear
    $ stor pwd
    s3://
    swift://

This also means that relative paths can be used. Relative paths are indicated
by omitting the ``//`` in the path and instead indicating a relative path, as
shown::

    $ stor cd s3://bucket/dir
    $ stor list s3:child
    s3://bucket/dir/child/file1
    s3://bucket/dir/child/file2
    $ stor list s3:./child
    s3://bucket/dir/child/file1
    s3://bucket/dir/child/file2
    $ stor list s3:..
    s3://bucket/a
    s3://bucket/b/obj1
    s3://bucket/dir/child/file1
    s3://bucket/dir/child/file2

``stdin`` and ``stdout``
------------------------

The CLI offers the ability to copy from ``stdin`` and output a path's
contents to ``stdout``.

To copy from ``stdin``, use the special ``-`` symbol. This means that the
user can pipe output from one command into the ``stor`` CLI::

    $ echo "hello world" | stor cp - s3://my/file1

The user can also output a path's contents to ``stdout`` using the ``cat``
subcommand::

    $ stor cat s3://my/file1
    hello world

Direct file transfer between OBS services or within one OBS service (server-side copy)
is not yet supported.
"""
import argparse
import copy
from functools import partial
import logging
import os
import shutil
import signal
import sys
import tempfile

import six
from six.moves import configparser

import stor
from stor import exceptions
from stor import settings
from stor import Path
from stor import utils

PRINT_CMDS = ('list', 'listdir', 'ls', 'cat', 'pwd', 'walkfiles')
SERVICES = ('s3', 'swift')

ENV_FILE = os.path.expanduser('~/.stor-cli.env')
PKG_ENV_FILE = os.path.join(os.path.dirname(__file__), 'default.env')


def perror(msg):
    """Print error message and exit."""
    sys.stderr.write(msg)
    sys.exit(1)


def force_exit(signum, frame):  # pragma: no cover
    sys.stderr.write(' Aborted\n')
    os._exit(1)

signal.signal(signal.SIGINT, force_exit)


class TempPath(Path):
    """Persist stdin to a temporary file for CLI operations with OBS."""
    def __del__(self):
        os.remove(str(self))


def _make_stdin_action(func, err_msg):
    """
    Return a StdinAction object that checks for stdin.
    func should be the function associated with the parser's -r flag
    that is not valid to use with stdin.
    """
    class StdinAction(argparse._StoreAction):
        def __call__(self, parser, namespace, values, option_string=None):
            if values == '-':
                if namespace.func == func:
                    raise argparse.ArgumentError(self, err_msg)
                else:
                    ntf = tempfile.NamedTemporaryFile(delete=False, mode='w')
                    try:
                        ntf.write(sys.stdin.read())
                    finally:
                        ntf.close()
                    super(StdinAction, self).__call__(parser,
                                                      namespace,
                                                      TempPath(ntf.name),
                                                      option_string=option_string)
            else:
                super(StdinAction, self).__call__(parser,
                                                  namespace,
                                                  values,
                                                  option_string=option_string)
    return StdinAction


def _get_env():
    """
    Get the current environment using the ENV_FILE.

    Returns a ConfigParser.
    """
    parser = configparser.SafeConfigParser()
    # if env file doesn't exist, copy over the package default
    if not os.path.exists(ENV_FILE):
        shutil.copyfile(PKG_ENV_FILE, ENV_FILE)
    with open(ENV_FILE) as fp:
        parser.readfp(fp)
    return parser


def _get_pwd(service=None):
    """
    Returns the present working directory for the given service,
    or all services if none specified.
    """
    parser = _get_env()
    if service:
        try:
            return utils.with_trailing_slash(parser.get('env', service))
        except configparser.NoOptionError as e:
            six.raise_from(ValueError('%s is an invalid service' % service), e)
    return [utils.with_trailing_slash(value) for name, value in parser.items('env')]


def _env_chdir(pth):
    """Sets the new current working directory."""
    parser = _get_env()
    if utils.is_obs_path(pth):
        service = Path(pth).drive.rstrip(':/')
    else:
        raise ValueError('%s is an invalid path' % pth)
    if pth != Path(pth).drive:
        if not Path(pth).isdir():
            raise ValueError('%s is not a directory' % pth)
        pth = utils.remove_trailing_slash(pth)
    parser.set('env', service, pth)
    with open(ENV_FILE, 'w') as outfile:
        parser.write(outfile)


def _clear_env(service=None):
    """Reset current working directory for the specified service or all if none specified."""
    parser = _get_env()
    if service:
        _env_chdir(service + '://')
    else:
        for name, value in parser.items('env'):
            _env_chdir(name + '://')


def _cat(pth):
    """Return the contents of a given path."""
    return pth.open().read()


def _obs_relpath_service(pth):
    """
    Check if path is an OBS relative path and if so,
    return the service, otherwise return an empty string.
    """
    prefixes = tuple(service + ':' for service in SERVICES)
    if pth.startswith(prefixes):
        if pth.startswith(tuple(p + '//' for p in prefixes)):
            return ''
        elif pth in prefixes or pth.startswith(tuple(p + '/' for p in prefixes)):
            raise ValueError('%s is an invalid path' % pth)
        parts = pth.split(':', 1)
        return parts[0]
    return ''


def get_path(pth, mode=None):
    """
    Convert string to a Path type.

    The string ``-`` is a special string depending on mode.
    With mode 'r', it represents stdin and a temporary file is created and returned.
    """
    service = _obs_relpath_service(pth)
    if not service:
        return Path(pth)

    relprefix = service + ':'

    pwd = Path(_get_pwd(service=service))
    if pwd == pwd.drive:
        raise ValueError('No current directory specified for relative path \'%s\'' % pth)

    pwd = utils.remove_trailing_slash(pwd)
    path_part = pth[len(relprefix):]
    split_parts = path_part.split('/')
    rel_part = split_parts[0]

    prefix = pwd
    depth = 1
    if rel_part == '..':
        # remove trailing slash otherwise we won't find the right parent
        prefix = utils.remove_trailing_slash(prefix)
        while len(split_parts) > depth and split_parts[depth] == '..':
            depth += 1
        if len(pwd[len(pwd.drive):].split('/')) > depth:
            for i in range(0, depth):
                prefix = prefix.parent
        else:
            raise ValueError('Relative path \'%s\' is invalid for current directory \'%s\''
                             % (pth, pwd))
    elif rel_part != '.':
        return prefix / path_part
    return prefix / path_part.split(rel_part, depth)[depth].lstrip('/')


def create_parser():
    parser = argparse.ArgumentParser(description='A command line interface for stor.')

    # todo: make default an environment variable?
    parser.add_argument('-c', '--config',
                        help='File containing configuration settings.',
                        type=str,
                        metavar='CONFIG_FILE')
    parser.add_argument('--version', help='Print version',
                        action='version',
                        version=stor.__version__)

    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True

    list_msg = 'List contents using the path as a prefix.'
    parser_list = subparsers.add_parser('list',
                                        help=list_msg,
                                        description=list_msg)
    parser_list.add_argument('path', type=get_path, metavar='PATH')
    parser_list.add_argument('-s', '--starts-with',
                             help='Append an additional path to the search path.',
                             type=str,
                             dest='starts_with',
                             metavar='PREFIX')
    parser_list.add_argument('-l', '--limit',
                             help='Limit the amount of results returned.',
                             type=int,
                             metavar='INT')
    parser_list.set_defaults(func=stor.list)

    ls_msg = 'List path as a directory.'
    parser_ls = subparsers.add_parser('ls',  # noqa
                                      help=ls_msg,
                                      description=ls_msg)
    parser_ls.add_argument('path', type=get_path, metavar='PATH')
    parser_ls.set_defaults(func=stor.listdir)

    cp_msg = 'Copy a source to a destination path.'
    cp_msg = 'Alias for copy.'
    parser_cp = subparsers.add_parser('cp',  # noqa
                                      help=cp_msg,
                                      description='%s\n \'-\' is a special character that allows'
                                                  ' for using stdin as the source.' % cp_msg)
    parser_cp.add_argument('-r',
                           help='Copy a directory and its subtree to the destination directory.'
                                ' Must be specified before any other flags.',
                           action='store_const',
                           dest='func',
                           const=stor.copytree,
                           default=stor.copy)
    parser_cp.add_argument('source',
                           type=get_path,
                           metavar='SOURCE',
                           action=_make_stdin_action(stor.copytree,
                                                     '- cannot be used with -r'))
    parser_cp.add_argument('dest', type=get_path, metavar='DEST')

    rm_msg = 'Remove file at a path.'
    parser_rm = subparsers.add_parser('rm',
                                      help=rm_msg,
                                      description='%s Use the -r flag to remove a tree.' % rm_msg)
    parser_rm.add_argument('-r',
                           help='Remove a path and all its contents.',
                           action='store_const',
                           dest='func',
                           const=stor.rmtree,
                           default=stor.remove)
    parser_rm.add_argument('path', type=get_path, metavar='PATH')

    walkfiles_msg = 'List all files under a path that match an optional pattern.'
    parser_walkfiles = subparsers.add_parser('walkfiles',
                                             help=walkfiles_msg,
                                             description=walkfiles_msg)
    parser_walkfiles.add_argument('-p', '--pattern',
                                  help='A regex pattern to match file names on.',
                                  type=str,
                                  metavar='REGEX')
    parser_walkfiles.add_argument('path', type=get_path, metavar='PATH')
    parser_walkfiles.set_defaults(func=stor.walkfiles)

    cat_msg = 'Output file contents to stdout.'
    parser_cat = subparsers.add_parser('cat', help=cat_msg, description=cat_msg)
    parser_cat.add_argument('path', type=partial(get_path, mode='r'), metavar='PATH')
    parser_cat.set_defaults(func=_cat)

    cd_msg = 'Change directory to a given OBS path.'
    parser_cd = subparsers.add_parser('cd', help=cd_msg, description=cd_msg)
    parser_cd.add_argument('path', type=get_path, metavar='PATH')
    parser_cd.set_defaults(func=_env_chdir)

    pwd_msg = 'Get the present working directory of a service or all current directories.'
    parser_pwd = subparsers.add_parser('pwd', help=pwd_msg, description=pwd_msg)
    parser_pwd.add_argument('service', nargs='?', type=str, metavar='SERVICE')
    parser_pwd.set_defaults(func=_get_pwd)

    clear_msg = 'Clear current directories of a specified service.'
    parser_clear = subparsers.add_parser('clear', help=clear_msg,
                                         description='%s The current directories of all services'
                                         ' will be cleared if SERVICE is omitted.' % clear_msg)
    parser_clear.add_argument('service', nargs='?', type=str, metavar='SERVICE')
    parser_clear.set_defaults(func=_clear_env)

    return parser


def process_args(args):
    args_copy = copy.copy(vars(args))
    config = args_copy.pop('config', None)
    func = args_copy.pop('func', None)
    pth = args_copy.pop('path', None)
    cmd = args_copy.pop('cmd', None)

    if config:
        settings.update(settings.parse_config_file(config))
    func_kwargs = {
        key: Path(val) if type(val) is TempPath else val
        for key, val in args_copy.items() if val
    }
    try:
        if pth:
            return func(pth, **func_kwargs)
        return func(**func_kwargs)
    except NotImplementedError:
        if pth:
            value = pth
        elif len(func_kwargs) > 0:
            value = list(func_kwargs.values())[0]
        else:
            perror('%s is not a valid command for the given input\n' % cmd)
        perror('%s is not a valid command for %s\n' % (cmd, value))
    except ValueError as exc:
        perror('Error: %s\n' % str(exc))
    except exceptions.RemoteError as exc:
        if type(exc) is exceptions.NotFoundError and pth:
            perror('Not Found: %s' % pth)
        perror('%s: %s\n' % (exc.__class__.__name__, str(exc)))


def print_results(results):
    if type(results) is str:
        sys.stdout.write(results)
        if not results.endswith('\n'):
            sys.stdout.write('\n')
    else:
        for result in results:
            sys.stdout.write('%s\n' % str(result))


def main():
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    s3_logger = logging.getLogger('stor.s3.progress')
    s3_logger.setLevel(logging.INFO)
    s3_logger.addHandler(handler)
    swift_logger = logging.getLogger('stor.swift.progress')
    swift_logger.setLevel(logging.INFO)
    swift_logger.addHandler(handler)

    settings._initialize()
    parser = create_parser()
    args = parser.parse_args()
    results = process_args(args)

    cmd = vars(args).get('cmd')
    if cmd in PRINT_CMDS:
        print_results(results)



"""
Provides functionality for accessing resources on Windows file systems.
"""
import ntpath
from stor import base


class WindowsPath(base.FileSystemPath):
    """Represents a windows path.

    This class inherits a vendored path.py ``Path`` object and
    overrides methods for cross compatibility with swift.

    This class overrides ``Path`` object's
    ``module`` attribute and sets it to ``ntpath`` to ensure
    that it always represents a windows path.
    """
    path_module = ntpath



"""
Provides utilities for accessing swift object storage.

Different configuration options are available at the module level, and
these variables are documented in the default configuration.

For swift authentication, the `auth_url`, `username`, and
`password` variables are used.

For methods that take conditions, the `initial_retry_sleep`,
``num_retries``, and `retry_sleep_function` variables are used to
configure the logic around retrying when the condition is not met.
Note that these variables can also be passed to the methods themselves.

Examples:

    A basic example of configuring the swift authentication parameters
    and downloading a directory::

    >>> from stor import swift
    >>> from stor import settings
    >>> settings.update({
    ...     'swift': {
    ...         'auth_url': 'swift_auth_url.com',
    ...         'username': 'swift_user',
    ...         'password': 'swift_pass'
    ...     }
    ... })
    >>> swift_path = swift.SwiftPath('swift://tenant/container/prefix')
    >>> swift_path.download('dest_dir')

More examples and documentations for swift methods can be found under
the `SwiftPath` class.
"""
import copy
from functools import partial
from functools import wraps
import json
import logging
import os
import tempfile
import threading

import six
from six.moves.urllib import parse
from swiftclient import exceptions as swift_exceptions
from swiftclient import service as swift_service
from swiftclient import client as swift_client
from swiftclient.utils import generate_temp_url

from stor import exceptions as stor_exceptions
from stor import is_swift_path
from stor import settings
from stor import utils
from stor.base import Path
from stor.obs import OBSFile
from stor.obs import OBSPath
from stor.obs import OBSUploadObject
from stor.posix import PosixPath
from stor.third_party.backoff import with_backoff


logger = logging.getLogger(__name__)
progress_logger = logging.getLogger('%s.progress' % __name__)

# python-swiftclient has a subtle bug in get_auth_keystone. If
# the auth_token is set beforehand and a token refresh happens,
# the invalid auth_token persists in the os_options argument,
# causing it to enter an infinite retry loop on authentication
# errors. Patch the get_auth_keystone function with one that
# clears the auth_token parameter on any Exceptions.
# Note that this behavior is tested in
# stor.tests.test_integration_swift:SwiftIntegrationTest.test_cached_auth_and_auth_invalidation
real_get_auth_keystone = swift_client.get_auth_keystone


def patched_get_auth_keystone(auth_url, user, key, os_options, **kwargs):
    try:
        return real_get_auth_keystone(auth_url, user, key, os_options, **kwargs)
    except:
        os_options.pop('auth_token', None)
        raise
swift_client.get_auth_keystone = patched_get_auth_keystone

# singleton that collects together auth tokens for storage URLs
_cached_auth_token_map = {}
_singleton_lock = threading.Lock()

# Content types that are assigned to empty directories
DIR_MARKER_TYPES = ('text/directory', 'application/directory')

# These variables are used to configure retry logic for swift.
# These variables can also be passed to the methods themselves
initial_retry_sleep = 1
"""The time to sleep before the first retry"""

# Make new Exceptions structure backwards compatible
SwiftError = stor_exceptions.RemoteError
NotFoundError = stor_exceptions.NotFoundError
ConditionNotMetError = stor_exceptions.ConditionNotMetError
SwiftFile = OBSFile
SwiftUploadObject = OBSUploadObject


def _default_retry_sleep_function(t, attempt):
    return t * 2

retry_sleep_function = _default_retry_sleep_function
"""The function that increases sleep time when retrying.

This function needs to take two integer
arguments (time slept last attempt, attempt number) and
return a time to sleep in seconds.
"""


def get_progress_logger():
    """Returns the swift progress logger"""
    return progress_logger


def _get_or_create_auth_credentials(tenant_name):
    """
    Gets the cached auth credential or creates one if none exists.

    If any auth setting is updated, all cached auth credentials are
    cleared and new auth credentials are created for the requested tenant.
    """
    options = settings.get()['swift']
    auth_url = options.get('auth_url')
    username = options.get('username')
    password = options.get('password')

    if tenant_name in _cached_auth_token_map:
        cached_params = _cached_auth_token_map[tenant_name]['params']
        if (auth_url == cached_params['auth_url'] and username == cached_params['username'] and
                password == cached_params['password']):
            return _cached_auth_token_map[tenant_name]['creds']
        else:
            _clear_cached_auth_credentials()

    storage_url, auth_token = swift_client.get_auth_keystone(
        auth_url, username, password,
        {'tenant_name': tenant_name},
    )
    creds = {
        'os_storage_url': storage_url,
        'os_auth_token': auth_token
    }
    cached_result = {
        'creds': creds,
        'params': {
            'auth_url': auth_url,
            'username': username,
            'password': password
        }
    }

    # Note: we are intentionally ignoring the rare race condition where
    # authentication starts in one thread, then settings are updated, and
    # then authentication finishes in the other.
    with _singleton_lock:
        _cached_auth_token_map[tenant_name] = cached_result
    return creds


def _clear_cached_auth_credentials():
    with _singleton_lock:
        _cached_auth_token_map.clear()


class UnavailableError(SwiftError):
    """Thrown when a 503 response is returned from swift"""
    pass


class FailedUploadError(UnavailableError):
    """Thrown when an upload fails because of availability issues"""
    pass


class UnauthorizedError(SwiftError):
    """Thrown when a 403 response is returned from swift.

    Note:
        Internal swift connection errors (e.g., when a particular node is
        unavailable) appear to translate themselves into 403 errors at the
        proxy layer, thus in general it's a good idea to retry on authorization
        errors as equivalent to unavailable errors when doing PUT or GET
        operations (list / stat / etc never hit this issue)."""
    pass


class AuthenticationError(SwiftError):
    """Thrown when a client has improper authentication credentials.

    Swiftclient throws this error when trying to authenticate with
    the keystone client. This is similar to a 401 HTTP response.
    """
    pass


class ConflictError(SwiftError):
    """Thrown when a 409 response is returned from swift

    This error is thrown when deleting a container and
    some object storage nodes report that the container
    has objects while others don't.
    """
    pass


class InconsistentDownloadError(SwiftError):
    """Thrown when an etag or content length does not match.

    Currently, we experience this during periods when the cluster is under
    heavy load, potentially because of unexpectedly quick terminations"""
    pass


class ConfigurationError(SwiftError):
    """Thrown when swift is not configured properly.

    Swift needs either module-level or environment authentication
    variables set in order to be configured.
    """
    pass


def _swift_retry(exceptions=None):
    """Allows `SwiftPath` methods to take optional retry configuration
    parameters for doing retry logic
    """
    def decorated(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = kwargs.pop('num_retries', settings.get()['swift']['num_retries'])
            initial_sleep = kwargs.pop('initial_retry_sleep',
                                       initial_retry_sleep)
            sleep_function = kwargs.pop('retry_sleep_function',
                                        retry_sleep_function)

            return with_backoff(func,
                                exceptions=exceptions,
                                sleep_function=sleep_function,
                                retries=retries,
                                initial_sleep=initial_sleep)(*args, **kwargs)
        return wrapper
    return decorated


def _swiftclient_error_to_descriptive_exception(exc):
    """Converts swiftclient errors to more descriptive exceptions with
    transaction ID"""
    # SwiftErrors catch Client exceptions and store them in the
    # 'exception' attribute. Try to get the client exception
    # here if there is one so that its http status can be
    # examined to throw specific exceptions.
    client_exception = getattr(exc, 'exception', exc)

    http_status = getattr(client_exception, 'http_status', None)
    exc_str = str(exc)
    exc_headers = getattr(client_exception, 'http_response_headers', None)
    if exc_headers and exc_headers.get('X-Trans-Id'):
        exc_str += ' X-Trans-Id: %s' % exc_headers['X-Trans-Id']
    if http_status == 403:
        logger.error('unauthorized error in swift operation - %s', exc_str)
        return UnauthorizedError(exc_str, exc)
    elif http_status == 404:
        return NotFoundError(exc_str, exc)
    elif http_status == 409:
        return ConflictError(exc_str, exc)
    elif http_status == 503:
        logger.error('unavailable error in swift operation - %s', exc_str)
        return UnavailableError(exc_str, exc)
    elif 'reset contents for reupload' in exc_str:
        # When experiencing HA issues, we sometimes encounter a
        # ClientException from swiftclient during upload. The exception
        # is thrown here -
        # https://github.com/openstack/python-swiftclient/blob/84d110c63ecf671377d4b2338060e9b00da44a4f/swiftclient/client.py#L1625  # nopep8
        # Treat this as a FailedUploadError
        logger.error('upload error in swift put_object operation - %s', exc_str)
        six.raise_from(FailedUploadError(exc_str, exc), exc)
    elif 'Unauthorized.' in exc_str:
        # Swiftclient catches keystone auth errors at
        # https://github.com/openstack/python-swiftclient/blob/master/swiftclient/client.py#L536 # nopep8
        # Parse the message since they don't bubble the exception or
        # provide more information
        logger.warning('auth error in swift operation - %s', exc_str)
        six.raise_from(AuthenticationError(exc_str, exc), exc)
    elif 'md5sum != etag' in exc_str or 'read_length != content_length' in exc_str:
        # We encounter this error when cluster is under heavy
        # replication load (at least that's the theory). So retry and
        # ensure we track consistency errors
        logger.error('Hit consistency issue. Likely related to'
                     ' cluster load: %s', exc_str)
        return InconsistentDownloadError(exc_str, exc)
    else:
        logger.error('unexpected swift error - %s', exc_str)
        return SwiftError(exc_str, exc)


def _propagate_swift_exceptions(func):
    """Bubbles all swift exceptions as `SwiftError` classes
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (swift_service.SwiftError,
                swift_exceptions.ClientException) as e:
            six.raise_from(_swiftclient_error_to_descriptive_exception(e), e)
    return wrapper


def _retry_on_cached_auth_err(func):
    """Retry a function with cleared auth credentials on AuthenticationError"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AuthenticationError:
            logger.info('auth failed, retrying with cleared auth cache')
            _clear_cached_auth_credentials()
            return func(*args, **kwargs)
    return wrapper


def _validate_manifest_upload(expected_objs, upload_results):
    """
    Given a list of expected object names and a list of dictionaries of
    `SwiftPath.upload` results, verify that all expected objects are in
    the upload results.
    """
    uploaded_objs = {
        r['object']
        for r in upload_results
        if r['success'] and r['action'] in ('upload_object', 'create_dir_marker')
    }
    return set(expected_objs).issubset(uploaded_objs)


def _validate_manifest_download(expected_objs, download_results):
    """
    Given a list of expected object names and a list of dictionaries of
    `SwiftPath.download` results, verify that all expected objects are in
    the download results.
    """
    downloaded_objs = {
        r['object']
        for r in download_results
        if r['success'] and r['action'] in ('download_object',)
    }
    return set(expected_objs).issubset(downloaded_objs)


class SwiftDownloadLogger(utils.BaseProgressLogger):
    def __init__(self):
        super(SwiftDownloadLogger, self).__init__(progress_logger)
        self.downloaded_bytes = 0

    def update_progress(self, result):
        """Tracks number of bytes downloaded.

        The ``read_length`` property in swift download results contains the
        total size of the object
        """
        self.downloaded_bytes += result.get('read_length', 0)

    def add_result(self, result):
        """Only add results to progress if they are ``download_object`` actions.

        The ``download_object`` action encompasses creating an empty directory
        marker
        """
        if result.get('action', None) == 'download_object':
            super(SwiftDownloadLogger, self).add_result(result)

    def get_start_message(self):
        return 'starting download'

    def get_finish_message(self):
        return 'download complete - %s' % self.get_progress_message()

    def get_progress_message(self):
        elapsed_time = self.get_elapsed_time()
        formatted_elapsed_time = self.format_time(elapsed_time)
        mb = self.downloaded_bytes / (1024 * 1024.0)
        mb_s = mb / elapsed_time.total_seconds() if elapsed_time else 0
        return (
            '%s\t'
            '%s\t'
            '%0.2f MB\t'
            '%0.2f MB/s'
        ) % (self.num_results, formatted_elapsed_time, mb, mb_s)


class SwiftUploadLogger(utils.BaseProgressLogger):
    def __init__(self, total_upload_objects, upload_object_sizes):
        super(SwiftUploadLogger, self).__init__(progress_logger)
        self.total_upload_objects = total_upload_objects
        self.upload_object_sizes = upload_object_sizes
        self.uploaded_bytes = 0

    def update_progress(self, result):
        """Keep track of total uploaded bytes by referencing the object sizes"""
        self.uploaded_bytes += self.upload_object_sizes.get(result['path'], 0)

    def add_result(self, result):
        """Only add results if they are ``upload_object`` and ``create_dir_marker``
        actions"""
        if result.get('action', None) in ('upload_object', 'create_dir_marker'):
            super(SwiftUploadLogger, self).add_result(result)

    def get_start_message(self):
        return 'starting upload of %s objects' % self.total_upload_objects

    def get_finish_message(self):
        return 'upload complete - %s' % self.get_progress_message()

    def get_progress_message(self):
        elapsed_time = self.get_elapsed_time()
        formatted_elapsed_time = self.format_time(elapsed_time)
        mb = self.uploaded_bytes / (1024 * 1024.0)
        mb_s = mb / elapsed_time.total_seconds() if elapsed_time else 0
        return (
            '%s/%s\t'
            '%s\t'
            '%0.2f MB\t'
            '%0.2f MB/s'
        ) % (self.num_results, self.total_upload_objects, formatted_elapsed_time, mb, mb_s)


class SwiftPath(OBSPath):
    """
    Provides the ability to manipulate and access resources on swift
    with a similar interface to the path library.
    """
    drive = 'swift://'

    def is_segment_container(self):
        """True if this path is a segment container"""
        container = self.container
        if not self.resource and container:
            return (container.startswith('.segments_') or
                    container.endswith('_segments') or
                    container.endswith('+segments'))
        else:
            return False

    @property
    def tenant(self):
        """Returns the tenant name from the path or return None"""
        parts = self._get_parts()
        return parts[0] if len(parts) > 0 and parts[0] else None

    @property
    def container(self):
        """Returns the container name from the path or None."""
        parts = self._get_parts()
        return parts[1] if len(parts) > 1 and parts[1] else None

    @property
    def resource(self):
        """Returns the resource as a ``PosixPath`` object or None.

        A resource can be a single object or a prefix to objects.
        Note that it's important to keep the trailing slash in a resource
        name for prefix queries.
        """
        parts = self._get_parts()
        joined_resource = '/'.join(parts[2:]) if len(parts) > 2 else None

        return self.parts_class(joined_resource) if joined_resource else None

    def _get_swift_connection_options(self, **options):
        """Returns options for constructing ``SwiftService`` and
        ``Connection`` objects.

        Args:
            options: Additional options that are directly passed
                into connection options.

        Raises:
            ConfigurationError: The needed swift environment variables
                aren't set.
        """
        global_options = settings.get()['swift']
        auth_url = global_options.get('auth_url')
        username = global_options.get('username')
        password = global_options.get('password')

        if not username or not password or not auth_url:
            raise ConfigurationError((
                'OS_AUTH_URL, OS_USERNAME, and OS_PASSWORD environment vars '
                'must be set for swift authentication. The username, password '
                'and auth_url settings variables may also be set with settings.update.'
            ))

        # Set additional options on top of what was passed in
        options['os_tenant_name'] = self.tenant
        options['os_auth_url'] = auth_url
        options['os_username'] = username
        options['os_password'] = password
        options.update(**_get_or_create_auth_credentials(self.tenant))

        # Merge options with global and local ones
        options = dict(swift_service._default_global_options,
                       **dict(swift_service._default_local_options,
                              **options))
        swift_service.process_options(options)
        return options

    def _get_swift_service(self, **options):
        """Initialize a swift service based on the path.

        Uses the tenant name of the path and an auth url to instantiate
        the swift service. The ``OS_AUTH_URL`` environment variable is used
        as the authentication url if set, otherwise the default auth
        url setting is used.

        Args:
            options: Additional options that are directly passed
                into swift service creation.

        Returns:
            swiftclient.service.SwiftService: The service instance.
        """
        conn_opts = self._get_swift_connection_options(**options)
        return swift_service.SwiftService(conn_opts)

    def _get_swift_connection(self, **options):
        """Initialize a swift client connection based on the path.

        The python-swiftclient package offers a couple ways to access data,
        with a raw Connection object being a lower-level interface to swift.
        For cases like reading individual objects, a raw Connection object
        is easier to utilize.

        Args:
            options: Additional options that are directly passed
                into swift connection creation.

        Returns:
            swiftclient.client.Connection: The connection instance.
        """
        conn_opts = self._get_swift_connection_options(**options)
        return swift_service.get_conn(conn_opts)

    @_retry_on_cached_auth_err
    @_propagate_swift_exceptions
    def _swift_connection_call(self, method_name, *args, **kwargs):
        """Instantiates a ``Connection`` object and runs ``method_name``.

        Note that obtaining the connection and doing the call in one method
        is intentional (instead of allowing the user to get a connection and
        then call a method on it directly). This is because sometimes a cached
        auth token can expire, causing a method to fail. If a method is retried,
        we want it to always get the swift connection again so that it will also
        re-auth in the case of an expired or invalid auth token.
        """
        connection = self._get_swift_connection()
        method = getattr(connection, method_name)
        return method(*args, **kwargs)

    @_retry_on_cached_auth_err
    @_propagate_swift_exceptions
    def _swift_service_call(self, method_name, *args, **kwargs):
        """Instantiates a ``SwiftService`` object and runs ``method_name``.

        Note that getting the swift service and doing the call in the same method
        is done for the same reasons explained in ``_swift_connection_call``.
        """
        method_options = copy.copy(kwargs)
        service_options = copy.deepcopy(method_options.pop('_service_options', {}))
        service_progress_logger = method_options.pop('_progress_logger', None)
        service = self._get_swift_service(**service_options)
        method = getattr(service, method_name)
        results_iter = method(*args, **method_options)

        results_iter = [results_iter] if isinstance(results_iter, dict) else results_iter

        results = []
        for r in results_iter:
            if 'error' in r:
                http_status = getattr(r['error'], 'http_status', None)
                if not http_status or http_status >= 400:
                    raise r['error']
            results.append(r)
            if service_progress_logger:
                service_progress_logger.add_result(r)

        return results

    @_swift_retry(exceptions=(NotFoundError, UnavailableError,
                              InconsistentDownloadError, UnauthorizedError))
    def read_object(self):
        """Reads an individual object.

        This method retries ``num_retries`` times if swift is unavailable or if
        the object is not found. View
        `module-level documentation <swiftretry>` for more
        information about configuring retry logic at the module or method
        level.
        """
        headers, content = self._swift_connection_call('get_object',
                                                       self.container,
                                                       self.resource)
        return content

    def temp_url(self, lifetime=300, method='GET', inline=True, filename=None):
        """Obtains a temporary URL to an object.

        Args:
            lifetime (int): The time (in seconds) the temporary
                URL will be valid
            method (str): The HTTP method that can be used on
                the temporary URL
            inline (bool, default True): If False, URL will have a
                Content-Disposition header that causes browser to download as
                attachment.
            filename (str, optional): A urlencoded filename to use for
                attachment, otherwise defaults to object name
        """
        global_options = settings.get()['swift']
        auth_url = global_options.get('auth_url')
        temp_url_key = global_options.get('temp_url_key')

        if not self.resource:
            raise ValueError('can only create temporary URL on object')
        if not temp_url_key:
            raise ValueError(
                'a temporary url key must be set with settings.update '
                'or by setting the OS_TEMP_URL_KEY environment variable')
        if not auth_url:
            raise ValueError(
                'an auth url must be set with settings.update '
                'or by setting the OS_AUTH_URL environment variable')

        obj_path = '/v1/%s' % self[len(self.drive):]
        # Generate the temp url using swifts helper. Note that this method is ONLY
        # useful for obtaining the temp_url_sig and the temp_url_expires parameters.
        # These parameters will be used to construct a properly-escaped temp url
        obj_url = generate_temp_url(obj_path, lifetime, temp_url_key, method)
        query_begin = obj_url.rfind('temp_url_sig', 0, len(obj_url))
        obj_url_query = obj_url[query_begin:]
        obj_url_query = dict(parse.parse_qsl(obj_url_query))

        query = ['temp_url_sig=%s' % obj_url_query['temp_url_sig'],
                 'temp_url_expires=%s' % obj_url_query['temp_url_expires']]
        if inline:
            query.append('inline')
        if filename:
            query.append('filename=%s' % parse.quote(filename))

        auth_url_parts = parse.urlparse(auth_url)
        return parse.urlunparse((auth_url_parts.scheme,
                                 auth_url_parts.netloc,
                                 parse.quote(obj_path),
                                 auth_url_parts.params,
                                 '&'.join(query),
                                 auth_url_parts.fragment))

    def write_object(self, content, **swift_upload_args):
        """Writes an individual object.

        Note that this method writes the provided content to a temporary
        file before uploading. This allows us to reuse code from swift's
        uploader (static large object support, etc.).

        For information about the retry logic of this method, view
        `SwiftPath.upload`.

        Args:
            content (str): The content of the object
            **swift_upload_args: Keyword arguments to pass to
                `SwiftPath.upload`
        """
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(content)
            fp.flush()
            suo = OBSUploadObject(fp.name, object_name=self.resource)
            return self.upload([suo], **swift_upload_args)

    def open(self, mode='r', swift_upload_options=None):
        """Opens a `SwiftFile` that can be read or written to.

        For examples of reading and writing opened objects, view
        `SwiftFile`.

        Args:
            mode (str): The mode of object IO. Currently supports reading
                ("r" or "rb") and writing ("w", "wb")
            swift_upload_options (dict): DEPRECATED (use `stor.settings.use()`
                instead). A dictionary of arguments that will be
                passed as keyword args to `SwiftPath.upload` if any writes
                occur on the opened resource.

        Returns:
            SwiftFile: The swift object.

        Raises:
            SwiftError: A swift client error occurred.
        """
        swift_upload_options = swift_upload_options or {}
        return SwiftFile(self, mode=mode, **swift_upload_options)

    @_swift_retry(exceptions=(ConditionNotMetError, UnavailableError))
    def list(self,
             starts_with=None,
             limit=None,
             condition=None,
             use_manifest=False,
             # intentionally not documented
             list_as_dir=False,
             ignore_segment_containers=True,
             ignore_dir_markers=False):
        """List contents using the resource of the path as a prefix.

        This method retries ``num_retries`` times if swift is unavailable
        or if the number of objects returned does not match the
        ``condition`` condition. View
        `module-level documentation <swiftretry>` for more
        information about configuring retry logic at the module or method
        level.

        Args:
            starts_with (str): Allows for an additional search path to
                be appended to the resource of the swift path. Note that the
                current resource path is treated as a directory
            limit (int): Limit the amount of results returned
            condition (function(results) -> bool): The method will only return
                when the results matches the condition.
            use_manifest (bool): Perform the list and use the data manfest file to validate
                the list.

        Returns:
            List[SwiftPath]: Every path in the listing.

        Raises:
            SwiftError: A swift client error occurred.
            ConditionNotMetError: Results were returned, but they did not
                meet the condition.
        """
        tenant = self.tenant
        prefix = self.resource
        full_listing = limit is None
        utils.validate_condition(condition)

        if use_manifest:
            object_names = utils.get_data_manifest_contents(self)
            manifest_cond = partial(utils.validate_manifest_list, object_names)
            condition = (utils.join_conditions(condition, manifest_cond)
                         if condition else manifest_cond)

        # When starts_with is provided, treat the resource as a
        # directory that has the starts_with parameter after it. This allows
        # the user to specify a path like tenant/container/mydir
        # and do an additional glob on a directory-like structure
        if starts_with:
            prefix = prefix / starts_with if prefix else starts_with

        list_kwargs = {
            'full_listing': full_listing,
            'limit': limit,
            'prefix': prefix
        }
        if self.container and list_as_dir:
            # Swift doesn't allow a delimeter for tenant-level listing,
            # however, this isn't a problem for list_as_dir since a tenant
            # will only have containers
            list_kwargs['delimiter'] = '/'

            # Ensure that the prefix has a '/' at the end of it for listdir
            list_kwargs['prefix'] = utils.with_trailing_slash(list_kwargs['prefix'])

        if self.container:
            results = self._swift_connection_call('get_container',
                                                  self.container,
                                                  **list_kwargs)
        else:
            results = self._swift_connection_call('get_account',
                                                  **list_kwargs)

        result_objs = results[1]
        if ignore_dir_markers:
            result_objs = [r for r in result_objs if r.get('content_type') not in DIR_MARKER_TYPES]

        path_pre = SwiftPath('%s%s' % (self.drive, tenant)) / (self.container or '')
        paths = list({
            path_pre / (r.get('name') or r['subdir'].rstrip('/'))
            for r in result_objs
        })

        if ignore_segment_containers:
            paths = [p for p in paths if not p.is_segment_container()]

        utils.check_condition(condition, paths)
        return paths

    def listdir(self, ignore_segment_containers=True):
        """Lists the path as a dir, returning top-level directories and files

        For information about retry logic on this method, see
        `SwiftPath.list`
        """
        return self.list(list_as_dir=True, ignore_segment_containers=ignore_segment_containers)

    @_swift_retry(exceptions=(ConditionNotMetError, UnavailableError))
    def glob(self, pattern, condition=None):
        """Globs all objects in the path with the pattern.

        This glob is only compatible with patterns that end in * because of
        swift's inability to do searches other than prefix queries.

        Note that this method assumes the current resource is a directory path
        and treats it as such. For example, if the user has a swift path of
        swift://tenant/container/my_dir (without the trailing slash), this
        method will perform a swift query with a prefix of mydir/pattern.

        This method retries ``num_retries`` times if swift is unavailable or if
        the number of globbed patterns does not match the ``condition``
        condition. View `module-level documentation <swiftretry>`
        for more information about configuring retry logic at the module or
        method level.

        Args:
            pattern (str): The pattern to match. The pattern can only have
                up to one '*' at the end.
            condition (function(results) -> bool): The method will only return
                when the number of results matches the condition.

        Returns:
            List[SwiftPath]: Every matching path.

        Raises:
            SwiftError: A swift client error occurred.
            ConditionNotMetError: Results were returned, but they did not
                meet the condition.
        """
        if pattern.count('*') > 1:
            raise ValueError('multiple pattern globs not supported')
        if '*' in pattern and not pattern.endswith('*'):
            raise ValueError('only prefix queries are supported')
        utils.validate_condition(condition)

        with settings.use({'swift': {'num_retries': 0}}):
            paths = self.list(starts_with=pattern.replace('*', ''))

        utils.check_condition(condition, paths)
        return paths

    @_swift_retry(exceptions=UnavailableError)
    def first(self):
        """Returns the first result from the list results of the path

        See `module-level retry <swiftretry>` documentation for more.

        Raises:
            SwiftError: A swift client error occurred.
        """
        with settings.use({'swift': {'num_retries': 0}}):
            results = self.list(limit=1)
        return results[0] if results else None

    @_swift_retry(exceptions=UnavailableError)
    def exists(self):
        """Checks existence of the path.

        Returns True if the path exists, False otherwise.

        See `module-level retry <swiftretry>` documentation for more.

        Returns:
            bool: True if the path exists, False otherwise.

        Raises:
            SwiftError: A non-404 swift client error occurred.
        """
        try:
            # first see if there is a specific corresponding object
            with settings.use({'swift': {'num_retries': 0}}):
                self.stat()
            return True
        except NotFoundError:
            pass
        try:
            # otherwise we could be a directory, so try to grab first
            # file/subfolder
            with settings.use({'swift': {'num_retries': 0}}):
                return bool(utils.with_trailing_slash(self).first())
        except NotFoundError:
            return False

    @_swift_retry(exceptions=(UnavailableError, InconsistentDownloadError,
                              UnauthorizedError))
    def download_object(self, out_file):
        """Downloads a single object to an output file.

        This method retries ``num_retries`` times if swift is unavailable.
        View module-level documentation for more information about configuring
        retry logic at the module or method level.

        Args:
            out_file (str): The output file

        Raises:
            ValueError: This method was called on a path that has no
                container or object
        """
        if not self.resource:
            raise ValueError('can only call download_object on object path')

        self._swift_service_call('download',
                                 container=self.container,
                                 objects=[self.resource],
                                 options={'out_file': out_file})

    @_swift_retry(exceptions=(UnavailableError, InconsistentDownloadError,
                              UnauthorizedError))
    def download_objects(self,
                         dest,
                         objects):
        """Downloads a list of objects to a destination folder.

        Note that this method takes a list of complete relative or absolute
        paths to objects (in contrast to taking a prefix). If any object
        does not exist, the call will fail with partially downloaded objects
        residing in the destination path.

        This method retries ``num_retries`` times if swift is unavailable.
        View `module-level documentation <swiftretry>` for more information
        about configuring retry logic at the module or method level.

        Args:
            dest (str): The destination folder to download to. The directory
                will be created if it doesnt exist.
            objects (List[str|PosixPath|SwiftPath]): The list of objects to
                download. The objects can paths relative to the download path
                or absolute swift paths. Any absolute swift path must be
                children of the download path

        Returns:
            dict: A mapping of all requested ``objs`` to their location on
                disk

        Raises:
            ValueError: This method was called on a path that has no
                container

        Examples:

            To download a objects to a ``dest/folder`` destination::

                from stor import path
                p = path('swift://tenant/container/dir/')
                results = p.download_objects('dest/folder', ['subdir/f1.txt',
                                                             'subdir/f2.txt'])
                print results
                {
                    'subdir/f1.txt': 'dest/folder/subdir/f1.txt',
                    'subdir/f2.txt': 'dest/folder/subdir/f2.txt'
                }

            To download full swift paths relative to a download path::

                from stor import path
                p = path('swift://tenant/container/dir/')
                results = p.download_objects('dest/folder', [
                    'swift://tenant/container/dir/subdir/f1.txt',
                    'swift://tenant/container/dir/subdir/f2.txt'
                ])
                print results
                {
                    'swift://tenant/container/dir/subdir/f1.txt': 'dest/folder/subdir/f1.txt',
                    'swift://tenant/container/dir/subdir/f2.txt': 'dest/folder/subdir/f2.txt'
                }
        """
        if not self.container:
            raise ValueError('cannot call download_objects on tenant with no container')

        # Convert requested download objects to full object paths
        obj_base = self.resource or PosixPath('')
        objs_to_download = {
            obj: SwiftPath(obj).resource if is_swift_path(obj) else obj_base / obj
            for obj in objects
        }

        for obj in objs_to_download:
            if is_swift_path(obj) and not obj.startswith(utils.with_trailing_slash(self)):
                raise ValueError(
                    '"%s" must be child of download path "%s"' % (obj, self))

        options = settings.get()['swift:download']

        service_options = {
            'object_dd_threads': options['object_threads'],
            'container_threads': options['container_threads']
        }
        download_options = {
            'prefix': utils.with_trailing_slash(self.resource),
            'out_directory': dest,
            'remove_prefix': True,
            'skip_identical': options['skip_identical'],
            'shuffle': options['shuffle']
        }
        results = self._swift_service_call('download',
                                           _service_options=service_options,
                                           container=self.container,
                                           objects=list(objs_to_download.values()),
                                           options=download_options)
        results = {r['object']: r['path'] for r in results}

        # Return results mapped back to their input name
        return {obj: results[objs_to_download[obj]] for obj in objects}

    @_swift_retry(exceptions=(ConditionNotMetError, UnavailableError,
                              InconsistentDownloadError))
    def download(self,
                 dest,
                 condition=None,
                 use_manifest=False):
        """Downloads a directory to a destination.

        This method retries ``num_retries`` times if swift is unavailable or if
        the returned download result does not match the ``condition``
        condition. View `module-level documentation <stor.swift>`
        for more information about configuring retry logic at the module or
        method level.

        Note that the destintation directory will be created automatically if
        it doesn't exist.

        Args:
            dest (str): The destination directory to download results to.
                The directory will be created if it doesn't exist.
            condition (function(results) -> bool): The method will only return
                when the results of download matches the condition. In the event of the
                condition never matching after retries, partially downloaded
                results will not be deleted. Note that users are not expected to write
                conditions for download without an understanding of the structure of the results.
            use_manifest (bool): Perform the download and use the data manfest file to validate
                the download.

        Raises:
            SwiftError: A swift client error occurred.
            ConditionNotMetError: Results were returned, but they did not
                meet the condition.

        Returns:
            List[dict]: A list of every operation performed in the download by the
                underlying swift client
        """
        if not self.container:
            raise ValueError('cannot call download on tenant with no container')
        utils.validate_condition(condition)

        if use_manifest:
            # Do a full list with the manifest before the download. This will retry until
            # all results in the manifest can be listed, which helps ensure the download
            # can be performed without having to be retried
            self.list(use_manifest=True)
            object_names = utils.get_data_manifest_contents(self)
            manifest_cond = partial(_validate_manifest_download, object_names)
            condition = (utils.join_conditions(condition, manifest_cond)
                         if condition else manifest_cond)

        options = settings.get()['swift:download']
        service_options = {
            'object_dd_threads': options['object_threads'],
            'container_threads': options['container_threads']
        }
        download_options = {
            'prefix': utils.with_trailing_slash(self.resource),
            'out_directory': dest,
            'remove_prefix': True,
            'skip_identical': options['skip_identical'],
            'shuffle': options['shuffle']
        }
        with SwiftDownloadLogger() as dl:
            results = self._swift_service_call('download',
                                               self.container,
                                               options=download_options,
                                               _progress_logger=dl,
                                               _service_options=service_options)

        utils.check_condition(condition, results)
        return results

    @_swift_retry(exceptions=(ConditionNotMetError, UnavailableError,
                              UnauthorizedError))
    def upload(self,
               to_upload,
               condition=None,
               use_manifest=False,
               headers=None):
        """Uploads a list of files and directories to swift.

        This method retries ``num_retries`` times if swift is unavailable or if
        the returned upload result does not match the ``condition``
        condition. View `module-level documentation <stor.swift>`
        for more information about configuring retry logic at the module or
        method level.

        For example::

            with path('/path/to/upload/dir'):
                path('swift://tenant/container').upload(['.'])

        Notes:

            - This method upload on paths relative to the current
              directory. In order to load files relative to a target directory,
              use path as a context manager to change the directory.

            - When large files are split into segments, they are uploaded
              to a segment container named .segments_${container_name}

        Args:
            to_upload (List): A list of file names, directory names, or
                OBSUploadObject objects to upload.
            condition (function(results) -> bool): The method will only return
                when the results of upload matches the condition. In the event of the
                condition never matching after retries, partially uploaded
                results will not be deleted. Note that users are not expected to write
                conditions for upload without an understanding of the structure of the results.
            use_manifest (bool): Generate a data manifest and validate the upload results
                are in the manifest. In case of a single directory being uploaded, the
                manifest file will be created inside this directory. For example::

                    stor.Path('swift://AUTH_foo/bar').upload(['logs'], use_manifest=True)

                The manifest will be located at
                ``swift://AUTH_foo/bar/logs/.data_manifest.csv``

                Alternatively, when multiple directories are uploaded, manifest file will
                be created in the current directory. For example::

                    stor.Path('swift://AUTH_foo/bar').upload(
                        ['logs', 'test.txt'], use_manifest=True)

                The manifest will be located at
                ``swift://AUTH_foo/bar/.data_manifest.csv``

            headers (List[str]): A list of object headers to apply to every object. Note
                that these are not applied if passing OBSUploadObjects directly to upload.
                Headers must be specified as a list of colon-delimited strings,
                e.g. ['X-Delete-After:1000']

        Raises:
            SwiftError: A swift client error occurred.
            ConditionNotMetError: The ``condition`` argument did not pass after ``num_retries``
                or the ``use_manifest`` option was turned on and the upload results could not
                be verified. Partially uploaded results are not deleted.

        Returns:
            List[dict]: A list of every operation performed in the upload by the
                underlying swift client
        """
        if not self.container:
            raise ValueError('must specify container when uploading')
        utils.validate_condition(condition)

        swift_upload_objects = [
            name for name in to_upload
            if isinstance(name, OBSUploadObject)
        ]
        all_files_to_upload = utils.walk_files_and_dirs([
            name for name in to_upload
            if not isinstance(name, OBSUploadObject)
        ])

        # Convert everything to swift upload objects and prepend the relative
        # resource directory to uploaded results. Ignore the manifest file in the case of
        # since it will be uploaded individually
        if use_manifest:
            if len(to_upload) == 1 and os.path.isdir(to_upload[0]):
                manifest_path_prefix = Path(to_upload[0])
            else:
                manifest_path_prefix = Path('.')
            manifest_file_name = manifest_path_prefix / utils.DATA_MANIFEST_FILE_NAME
        else:
            manifest_path_prefix = None
            manifest_file_name = None
        resource_base = utils.with_trailing_slash(self.resource) or PosixPath('')
        upload_object_options = {'header': headers or []}
        swift_upload_objects.extend([
            OBSUploadObject(f,
                            object_name=resource_base / utils.file_name_to_object_name(f),
                            options=upload_object_options)
            for f in all_files_to_upload if f != manifest_file_name
        ])

        if use_manifest:
            # Generate the data manifest and save it remotely
            object_names = [o.object_name for o in swift_upload_objects]
            utils.generate_and_save_data_manifest(manifest_path_prefix, object_names)
            manifest_obj_name = resource_base / utils.file_name_to_object_name(manifest_file_name)
            manifest_obj = OBSUploadObject(manifest_file_name,
                                           object_name=manifest_obj_name,
                                           options=upload_object_options)
            self._swift_service_call('upload', self.container, [manifest_obj])

            # Make a condition for validating the upload
            manifest_cond = partial(_validate_manifest_upload, object_names)
            condition = (utils.join_conditions(condition, manifest_cond)
                         if condition else manifest_cond)

        options = settings.get()['swift:upload']
        service_options = {
            'object_uu_threads': options['object_threads'],
            'segment_threads': options['segment_threads']
        }
        upload_options = {
            'segment_size': options['segment_size'],
            'use_slo': options['use_slo'],
            'segment_container': '.segments_%s' % self.container,
            'leave_segments': options['leave_segments'],
            'changed': options['changed'],
            'skip_identical': options['skip_identical'],
            'checksum': options['checksum']
        }
        with SwiftUploadLogger(len(swift_upload_objects), all_files_to_upload) as ul:
            results = self._swift_service_call('upload',
                                               self.container,
                                               swift_upload_objects,
                                               options=upload_options,
                                               _progress_logger=ul,
                                               _service_options=service_options)

        utils.check_condition(condition, results)
        return results

    @_swift_retry(exceptions=(UnavailableError, UnauthorizedError))
    def remove(self):
        """Removes a single object.

        This method retries ``num_retries`` times if swift is unavailable.
        View `module-level documentation <swiftretry>` for more
        information about configuring retry logic at the module or method
        level.

        Raises:
            ValueError: The path is invalid.
            SwiftError: A swift client error occurred.
        """
        if not self.container or not self.resource:
            raise ValueError('path must contain a container and resource to '
                             'remove a single file')

        return self._swift_service_call('delete',
                                        self.container,
                                        [self.resource])

    @_swift_retry(exceptions=(UnavailableError, ConflictError,
                              ConditionNotMetError, UnauthorizedError))
    def rmtree(self):
        """Removes a resource and all of its contents.
        This method retries ``num_retries`` times if swift is unavailable.
        View `module-level documentation <stor.swift>` for more
        information about configuring retry logic at the module or method
        level.

        Note that when removing a container, the associated segment container
        will also be removed if it exists. So, if one removes
        ``swift://tenant/container``, ``swift://tenant/container_segments``
        will also be deleted.

        Note:
            Calling rmtree on a directory marker will delete everything under the
            directory marker but not the marker itself.

        Raises:
            SwiftError: A swift client error occurred.
            ConditionNotMetError: Listing the objects after rmtree returns results,
                indicating something went wrong with the rmtree operation
        """
        if not self.container:
            raise ValueError('swift path must include container for rmtree')

        # Ensure that we treat this path as a dir
        to_delete = utils.with_trailing_slash(self)

        deleting_segments = 'segments' in self.container
        if deleting_segments:
            logger.warning('performing rmtree with segment container "%s". '
                           'This could cause issues when accessing objects '
                           'referencing those segments. Note that segments '
                           'and segment containers are automatically deleted '
                           'when their associated objects or containers are '
                           'deleted.', self.container)

        options = settings.get()

        service_options = {
            'object_dd_threads': options['swift:delete']['object_threads']
        }

        def _ignore_not_found(service_call):
            """Ignores 404 errors when performing a swift service call"""
            def wrapper(*args, **kwargs):
                try:
                    return service_call(*args, **kwargs)
                except NotFoundError:
                    return []
            return wrapper

        if not to_delete.resource:
            results = _ignore_not_found(self._swift_service_call)('delete',
                                                                  to_delete.container,
                                                                  _service_options=service_options)
            # Try to delete a segment container since swift client does not
            # do this automatically
            if not deleting_segments:
                segment_containers = ('%s_segments' % to_delete.container,
                                      '.segments_%s' % to_delete.container,
                                      '%s+segments' % to_delete.container)
                for segment_container in segment_containers:
                    _ignore_not_found(self._swift_service_call)('delete',
                                                                segment_container,
                                                                _service_options=service_options)
        else:
            objs_to_delete = [p.resource for p in to_delete.list()]
            results = _ignore_not_found(self._swift_service_call)('delete',
                                                                  self.container,
                                                                  objs_to_delete,
                                                                  _service_options=service_options)

        # Verify that all objects have been deleted before returning. Otherwise try deleting again
        with settings.use({'swift': {'num_retries': 0}}):
            _ignore_not_found(to_delete.list)(condition=lambda results: len(results) == 0)

        return results

    @_swift_retry(exceptions=(UnavailableError, UnauthorizedError))
    def remove_container(self):
        """
        Remove swift container if it's not empty.
        """
        if not self.container:
            raise ValueError('swift path must include container for remove_container')
        if self.resource:
            raise ValueError('swift path must not include resource for remove_container')

        return self._swift_connection_call('delete_container', self.container)

    @_swift_retry(exceptions=UnavailableError)
    def stat(self):
        """Performs a stat on the path.

        Note that the path can be a tenant, container, or
        object. Using ``stat`` on a directory of objects will
        produce a `NotFoundError`.

        This method retries ``num_retries`` times if swift is unavailable.
        View `module-level documentation <swiftretry>` for more information
        about configuring retry logic at the module or method level.

        The return value is dependent on if the path points to a tenant,
        container, or object.

        For tenants, an example return dictionary is the following::

            {
                'headers': {
                    'content-length': u'0',
                    'x-account-storage-policy-3xreplica-container-count': u'4655',
                    'x-account-meta-temp-url-key': 'temp_url_key',
                    'x-account-object-count': '428634',
                    'x-timestamp': '1450732238.21562',
                    'x-account-storage-policy-3xreplica-bytes-used': '175654877752',
                    'x-trans-id': u'transaction_id',
                    'date': u'Wed, 06 Apr 2016 18:30:28 GMT',
                    'x-account-bytes-used': '175654877752',
                    'x-account-container-count': '4655',
                    'content-type': 'text/plain; charset=utf-8',
                    'accept-ranges': 'bytes',
                    'x-account-storage-policy-3xreplica-object-count': '428634'
                },
                'Account': 'AUTH_seq_upload_prod',
                # The number of containers in the tenant
                'Containers': 31,
                # The number of objects in the tenant
                'Objects': '19955615',
                # The total bytes used in the tenant
                'Bytes': '24890576770484',
                'Containers-in-policy-"3xreplica"': '31',
                'Objects-in-policy-"3xreplica"': '19955615',
                'Bytes-in-policy-"3xreplica"': '24890576770484',
                # The tenant ACLs. An empty dict is returned if the user
                # does not have admin privileges on the tenant
                'Account-Access': {
                    'admin': ['swft_labprod_admin'],
                    'read-only': ['seq_upload_rnd','swft_labprod'],
                    'read-write': ['svc_svc_seq']
                }
            }

        For containers, an example return dictionary is the following::

            {
                'headers': {
                    'content-length': '0',
                    'x-container-object-count': '99',
                    'accept-ranges': 'bytes',
                    'x-storage-policy': '3xReplica',
                    'date': 'Wed, 06 Apr 2016 18:32:47 GMT',
                    'x-timestamp': '1457631707.95682',
                    'x-trans-id': 'transaction_id',
                    'x-container-bytes-used': '5389060',
                    'content-type': 'text/plain; charset=utf-8'
                },
                'Account': 'AUTH_seq_upload_prod',
                'Container': '2016-01',
                # The number of objects in the container
                'Objects': '43868',
                # The size of all objects in the container
                'Bytes': '55841489571',
                # Read and write ACLs for the container
                'Read-ACL': '',
                'Write-ACL': '',
                'Sync-To': '',
                'Sync-Key': ''
            }

        For objects, an example return dictionary is the following::

            {
                'headers': {
                    'content-length': '0',
                    'x-delete-at': '1459967915',
                    'accept-ranges': 'bytes',
                    'last-modified': 'Wed, 06 Apr 2016 18:21:56 GMT',
                    'etag': 'd41d8cd98f00b204e9800998ecf8427e',
                    'x-timestamp': '1459966915.96956',
                    'x-trans-id': 'transaction_id',
                    'date': 'Wed, 06 Apr 2016 18:33:48 GMT',
                    'content-type': 'text/plain',
                    'x-object-meta-mtime': '1459965986.000000'
                },
                'Account': 'AUTH_seq_upload_prod',
                'Container': '2016-01',
                'Object': PosixPath('object.txt'),
                'Content-Type': 'application/octet-stream',
                # The size of the object
                'Content-Length': '112',
                # The last time the object was modified
                'Last-Modified': 'Fri, 15 Jan 2016 05:22:46 GMT',
                # The MD5 checksum of the object. NOTE that if a large
                # object is uploaded in segments that this will be the
                # checksum of the *manifest* file of the object
                'ETag': '87f0b7f04557315e6d1e6db21742d31c',
                'Manifest': None
            }

        Raises:
            NotFoundError: When the tenant, container, or
                object can't be found.
        """
        stat_objects = [self.resource] if self.resource else None
        result = self._swift_service_call('stat',
                                          container=self.container,
                                          objects=stat_objects)[0]

        stat_values = {
            k.replace(' ', '-'): v
            for k, v in result['items']
        }
        stat_values['headers'] = result['headers']

        if result['action'] == 'stat_account':
            # Load account ACLs
            stat_values['Access-Control'] = json.loads(
                result['headers'].get('x-account-access-control', '{}'))

        return stat_values

    def getsize(self):
        """Returns content-length of object in Swift.

        Note that for containers / tenants, there will be no content-length, in
        which case this function returns 0 (``os.path.getsize`` has no
        contract)"""
        return int(self.stat().get('Content-Length', 0))

    @_swift_retry(exceptions=(UnavailableError, UnauthorizedError))
    def post(self, options=None):
        """Post operations on the path.

        This method retries ``num_retries`` times if swift is unavailable.
        View `module-level documentation <swiftretry>` for more
        information about configuring retry logic at the module or method
        level.

        Args:
            options (dict): A dictionary containing options to override the
                global options specified during the service object creation.
                These options are applied to all post operations performed by
                this call, unless overridden on a per object basis. Possible
                options are given below::

                    {
                        'meta': [],   # Meta values will be prefixed with X-Object-Meta
                                      # (or X-Container*  X-Account* depending on path type)
                        'header': [], # Header values will not be manipulated like meta values
                                      # when being posted.
                        'read_acl': None,   # For containers only
                        'write_acl': None,  # For containers only
                        'sync_to': None,    # For containers only
                        'sync_key': None    # For containers only
                    }

        Raises:
            SwiftError: A swift client error occurred.
        """
        return self._swift_service_call('post',
                                        container=self.container,
                                        objects=[self.resource] if self.resource else None,
                                        options=options)

    def _noop(attr_name):
        def wrapper(self):
            return type(self)(self)
        wrapper.__name__ = attr_name
        wrapper.__doc__ = 'No-op for %r' % attr_name
        return wrapper

    abspath = _noop('abspath')
    realpath = _noop('realpath')
    expanduser = _noop('expanduser')

    def isdir(self):
        if not self.resource:
            return self.exists()
        try:
            if utils.with_trailing_slash(self).first():
                return True
        except NotFoundError:
            pass
        try:
            return 'directory' in self.stat().get('Content-Type', '')
        except NotFoundError:
            pass
        return False

    def isfile(self):
        """Checks the object exists & is not a directory sentinel on Swift.
        """
        try:
            return self.resource and 'directory' not in self.stat().get('Content-Type', '')
        except NotFoundError:
            return False

    @_swift_retry(exceptions=UnavailableError)
    def walkfiles(self, pattern=None):
        """Iterates over listed files that match an optional pattern.

        Args:
            pattern (str, optional): Only return files that match this pattern

        Returns:
            Iter[SwiftPath]: All files that match the optional pattern. Swift directory
                markers are not returned.
        """
        with settings.use({'swift': {'num_retries': 0}}):
            for f in self.list(ignore_dir_markers=True):
                if pattern is None or f.fnmatch(pattern):
                    yield f



import posixpath
import sys

from cached_property import cached_property
import six
from swiftclient.service import SwiftError
from swiftclient.service import SwiftUploadObject

from stor.base import Path
from stor.posix import PosixPath
from stor import utils


def _delegate_to_buffer(attr_name, valid_modes=None):
    """Factory function that delegates file-like properties to underlying buffer"""
    def wrapper(self, *args, **kwargs):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        if valid_modes and self.mode not in valid_modes:
            raise TypeError('File must be in modes %s to %r' %
                            (valid_modes, attr_name))
        func = getattr(self._buffer, attr_name)
        return func(*args, **kwargs)
    wrapper.__name__ = attr_name
    wrapper.__doc__ = getattr(six.BytesIO(), attr_name).__doc__
    return wrapper


class OBSUploadObject(SwiftUploadObject):
    """
    An upload object similar to swiftclient's SwiftUploadObject that allows the user
    to specify a destination file name (full key) and upload options.
    """
    def __init__(self, source, object_name, options=None):
        """
        An OBSUploadObject must be initialized with a source and destination path.

        Args:
            source (str): A path that specifies a source file.
            dest (str): A path that specifies a destination file name (full key)
        """
        try:
            super(OBSUploadObject, self).__init__(source, object_name=object_name, options=options)
        except SwiftError as exc:
            if 'SwiftUploadObject' in exc.value:
                msg = exc.value.replace('SwiftUploadObject', 'OBSUploadObject', 1)
            else:
                msg = exc.value
            raise ValueError(msg)


class OBSPath(Path):
    """
    A base class that defines all methods available from Path objects that
    interface with remote services, such as Swift and S3.
    """
    path_module = posixpath
    parts_class = PosixPath

    def __init__(self, pth):
        """
        Validates OBS path is in the proper format.

        Args:
            pth (str): A path that matches the format of
                ``(drive)://{bucket_name}/{rest_of_path}`` where (drive)
                can be any of s3 or swift.
                The ``(drive)://`` prefix is required in the path.
        """
        if not hasattr(pth, 'startswith') or not pth.startswith(self.drive):
            raise ValueError('path must have %s (got %r)' % (self.drive, pth))
        return super(OBSPath, self).__init__(pth)

    copy = utils.copy
    copytree = utils.copytree

    def dirname(self):
        """
        Return the directory name of self.
        """
        assert self.startswith(self.drive)
        rest = self[len(self.drive):]
        return self.path_class(self.drive + self.path_module.dirname(rest))

    def __repr__(self):
        return '%s("%s")' % (type(self).__name__, self)

    def is_ambiguous(self):
        """Returns true if it cannot be determined if the path is a
        file or directory
        """
        return not self.endswith('/') and not self.ext

    def _get_parts(self):
        """Returns the path parts (excluding the drive) as a list of strings."""
        if len(self) > len(self.drive):
            return self[len(self.drive):].split('/')
        else:
            return []

    @property
    def name(self):
        """The name of the path, mimicking path.py's name property"""
        return self.parts_class(super(OBSPath, self).name)

    @property
    def parent(self):
        """The parent of the path, mimicking path.py's parent property"""
        return self.path_class(super(OBSPath, self).parent)

    @property
    def resource(self):
        """Returns the resource as a ``PosixPath`` object or None.

        A resource can be a single object or a prefix to objects.
        Note that it's important to keep the trailing slash in a resource
        name for prefix queries.
        """
        parts = self._get_parts()
        joined_resource = '/'.join(parts[1:]) if len(parts) > 1 else None

        return self.parts_class(joined_resource) if joined_resource else None

    def normpath(self):
        """Normalize path following linux conventions (keeps drive prefix)"""
        normed = posixpath.normpath('/' + str(self)[len(self.drive):])[1:]
        return self.path_class(self.drive + normed)

    def read_object(self):
        raise NotImplementedError

    def write_object(self, content):
        """Writes an individual object."""
        raise NotImplementedError

    def open(self, mode='r'):
        """
        Opens a OBSFile that can be read or written to and is uploaded to
        the remote service.
        """
        raise NotImplementedError

    def list(self):
        """List contents using the resource of the path as a prefix."""
        raise NotImplementedError

    def listdir(self):
        """list the path as a dir, returning top-level directories and files."""
        raise NotImplementedError

    def glob(self, pattern):
        """ Glob for pattern relative to this directory.

        Note that Swift currently only supports a single trailing *"""
        raise NotImplementedError

    def exists(self):
        """Checks whether path exists on local filesystem or on swift.

        For directories on swift, checks whether directory sentinel exists or
        at least one subdirectory exists"""
        raise NotImplementedError

    def isabs(self):
        return True

    def isdir(self):
        raise NotImplementedError

    def isfile(self):
        raise NotImplementedError

    def islink(self):
        return False

    def ismount(self):
        return True

    def getsize(self):
        """Returns size, in bytes of path."""
        raise NotImplementedError

    def remove(self):
        """ Delete single path or object """
        raise NotImplementedError

    def rmtree(self):
        """Delete entire directory (or all paths starting with prefix).

        See shutil.rmtree"""
        raise NotImplementedError

    def stat(self):
        raise NotImplementedError

    def walkfiles(self, pattern=None):
        """Iterate over files recursively.

        Args:
            pattern (str, optional): Limits the results to files
                with names that match the pattern.  For example,
                ``mydir.walkfiles('*.tmp')`` yields only files with the ``.tmp``
                extension.

        Returns:
            Iter[Path]: Files recursively under the path
        """
        for f in self.list(ignore_dir_markers=True):
            if pattern is None or f.fnmatch(pattern):
                yield f

    def download_object(self, dest):
        """Download a single path or object to file."""
        raise NotImplementedError

    def download(self, dest, condition=None, use_manifest=False, **kwargs):
        """Download a directory."""
        raise NotImplementedError

    def upload(self, source, condition=None, use_manifest=False, headers=None):
        """Upload a list of files and directories to a directory."""
        raise NotImplementedError


class OBSFile(object):
    """
    Provides methods for reading and writing OBS objects returned by
    `OBSPath.open`.

    Objects are retrieved from `OBSPath.open`. For example::

        obj = Path('swift://tenant/container/object').open(mode='r')
        contents = obj.read()

    The above opens an object and reads its contents. To write to an
    object::

        obj = Path('s3://bucket/object').open(mode='w')
        obj.write('hello ')
        obj.write('world')
        obj.close()

    Note that the writes will not be commited until the object has been
    closed. It is recommended to use `OBSPath.open` as a context manager
    to avoid forgetting to close the resource::

        with Path('swift://tenant/container/object').open(mode='r') as obj:
            obj.write('Hello world!')
    """
    closed = False
    _READ_MODES = ('r', 'rb')
    _WRITE_MODES = ('w', 'wb')
    _VALID_MODES = _READ_MODES + _WRITE_MODES

    def __init__(self, pth, mode='r', **kwargs):
        """Initializes a file object

        Args:
            pth (Path): The path that represents an individual
                object
            mode (str): The mode of the resource. Can be "r" and "rb" for
                reading the resource and "w" and "wb" for writing the
                resource.
        """
        if mode not in self._VALID_MODES:
            raise ValueError('invalid mode for file: %r' % mode)
        self._path = pth
        self.mode = mode
        self._kwargs = kwargs

    def __enter__(self):
        if self.closed:
            raise ValueError('I/O operation on closed file.')
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __iter__(self):
        # ape file object behavior by returning self (and attempting *not* to
        # give users access to underlying buffer)
        return self

    def __del__(self):
        self.close()

    @property
    def stream_cls(self):
        """The class used for the IO stream"""
        return six.BytesIO if self.mode in ('rb', 'wb') else six.StringIO

    @cached_property
    def _buffer(self):
        "Cached buffer of data read from or to be written to Object Storage"
        if self.mode in ('r', 'rb'):
            return self.stream_cls(self._path.read_object())
        elif self.mode in ('w', 'wb'):
            return self.stream_cls()
        else:
            raise ValueError('cannot obtain buffer in mode: %r' % self.mode)

    seek = _delegate_to_buffer('seek', valid_modes=_VALID_MODES)
    tell = _delegate_to_buffer('tell', valid_modes=_VALID_MODES)

    read = _delegate_to_buffer('read', valid_modes=_READ_MODES)
    readlines = _delegate_to_buffer('readlines', valid_modes=_READ_MODES)
    readline = _delegate_to_buffer('readline', valid_modes=_READ_MODES)

    # In Python 3 it's __next__, in Python 2 it's next()
    #
    # TODO: Only use in Python 2 context
    if sys.version_info >= (3, 0):
        __next__ = _delegate_to_buffer('__next__', valid_modes=_READ_MODES)  # pragma: no cover
    else:
        next = _delegate_to_buffer('next', valid_modes=_READ_MODES)  # pragma: no cover

    write = _delegate_to_buffer('write', valid_modes=_WRITE_MODES)
    writelines = _delegate_to_buffer('writelines', valid_modes=_WRITE_MODES)
    truncate = _delegate_to_buffer('truncate', valid_modes=_WRITE_MODES)

    @property
    def name(self):
        return self._path

    def close(self):
        if self.closed:
            return
        if self.mode in self._WRITE_MODES:
            self.flush()
        self._buffer.close()
        self.closed = True
        del self.__dict__['_buffer']

    def flush(self):
        """Flushes the write buffer to the OBS path (if it exists)"""
        if self.mode not in self._WRITE_MODES:
            raise TypeError("File must be in modes %s to 'flush'" %
                            (self._WRITE_MODES,))
        if self._buffer.tell():
            self._path.write_object(self._buffer.getvalue(),
                                    **self._kwargs)
